from typing import List, Union
from math import sqrt, pi, log10

import torch
from torch import Tensor

__all__ = [
    "get_reversed_params",
    "get_density_profiles",
    "uniform_sampler",
    "logdist_sampler",
    "triangular_sampler",
    "get_param_labels",
    "get_d_rhos",
    "get_slds_from_d_rhos",
]


def uniform_sampler(low: Union[float, Tensor], high: Union[float, Tensor], *shape, device=None, dtype=None):
    if isinstance(low, Tensor):
        device, dtype = low.device, low.dtype
    return torch.rand(*shape, device=device, dtype=dtype) * (high - low) + low


def logdist_sampler(low: Union[float, Tensor], high: Union[float, Tensor], *shape, device=None, dtype=None):
    if isinstance(low, Tensor):
        device, dtype = low.device, low.dtype
        low, high = map(torch.log10, (low, high))
    else:
        low, high = map(log10, (low, high))
    return 10 ** (torch.rand(*shape, device=device, dtype=dtype) * (high - low) + low)


def triangular_sampler(low: Union[float, Tensor], high: Union[float, Tensor], *shape, device=None, dtype=None):
    if isinstance(low, Tensor):
        device, dtype = low.device, low.dtype

    x = torch.rand(*shape, device=device, dtype=dtype)

    return (high - low) * (1 - torch.sqrt(x)) + low


def get_reversed_params(thicknesses: Tensor, roughnesses: Tensor, slds: Tensor):
    reversed_slds = torch.cumsum(
        torch.flip(
            torch.diff(
                torch.cat([torch.zeros(slds.shape[0], 1).to(slds), slds], dim=-1),
                dim=-1
            ), (-1,)
        ),
        dim=-1
    )
    reversed_thicknesses = torch.flip(thicknesses, [-1])
    reversed_roughnesses = torch.flip(roughnesses, [-1])
    reversed_params = torch.cat([reversed_thicknesses, reversed_roughnesses, reversed_slds], -1)

    return reversed_params


def get_density_profiles_sld(
        thicknesses: Tensor,
        roughnesses: Tensor,
        slds: Tensor,
        z_axis: Tensor = None,
        num: int = 1000
):
    """Generates SLD profiles (and their derivative) based on batches of thicknesses, roughnesses and layer SLDs. 
    
    The axis has its zero at the top (ambient medium) interface and is positive inside the film.

    Args:
        thicknesses (Tensor): the layer thicknesses (top to bottom)
        roughnesses (Tensor): the interlayer roughnesses (top to bottom)
        slds (Tensor): the layer SLDs (top to bottom)
        z_axis (Tensor, optional): a custom depth (z) axis. Defaults to None.
        num (int, optional): number of discretization points for the profile. Defaults to 1000.

    Returns:
        tuple: the z axis, the computed density profile rho(z) and the derivative of the density profile drho/dz(z)
    """
    assert torch.all(roughnesses >= 0), 'Negative roughness happened'
    assert torch.all(thicknesses >= 0), 'Negative thickness happened'

    sample_num = thicknesses.shape[0]

    d_rhos = get_d_rhos(slds)

    zs = torch.cumsum(torch.cat([torch.zeros(sample_num, 1).to(thicknesses), thicknesses], dim=-1), dim=-1)

    if z_axis is None:
        z_axis = torch.linspace(- zs.max() * 0.1, zs.max() * 1.1, num, device=thicknesses.device)[None]
    elif len(z_axis.shape) == 1:
        z_axis = z_axis[None]

    sigmas = roughnesses * sqrt(2)

    profile = get_erf(z_axis[:, None], zs[..., None], sigmas[..., None], d_rhos[..., None]).sum(1)

    d_profile = get_gauss(z_axis[:, None], zs[..., None], sigmas[..., None], d_rhos[..., None]).sum(1)

    z_axis = z_axis[0]

    return z_axis, profile, d_profile


def get_d_rhos(slds: Tensor) -> Tensor:
    d_rhos = torch.cat([slds[:, 0][:, None], torch.diff(slds, dim=-1)], -1)
    return d_rhos


def get_slds_from_d_rhos(d_rhos: Tensor) -> Tensor:
    slds = torch.cumsum(d_rhos, dim=-1)
    return slds


def get_erf(z, z0, sigma, amp):
    return (torch.erf((z - z0) / sigma) + 1) * amp / 2


def get_gauss(z, z0, sigma, amp):
    return amp / (sigma * sqrt(2 * pi)) * torch.exp(- (z - z0) ** 2 / 2 / sigma ** 2)

def get_density_profiles(
    thicknesses: torch.Tensor,
    roughnesses: torch.Tensor,
    slds: torch.Tensor,
    ambient_sld: torch.Tensor = None,
    z_axis: torch.Tensor = None,
    num: int = 1000,
    padding_left: float = 0.2,
    padding_right: float = 1.1,
):
    """
    Args:
        thicknesses (Tensor): finite layer thicknesses.
        roughnesses (Tensor): interface roughnesses for all transitions (ambient→layer1 ... layerN→substrate).
        slds (Tensor): SLDs for the finite layers + substrate.
        ambient_sld (Tensor, optional): SLD for the top ambient. Defaults to 0.0 if None.
        z_axis (Tensor, optional): a custom depth axis. If None, a linear axis is generated.
        num (int): number of points in the generated z-axis (if z_axis is None).
        padding_left (float): factor to extend the negative (above the surface) portion of z-axis.
        padding_right (float): factor to extend the positive (into the sample) portion of z-axis.

    Returns:
        (z_axis, profile, d_profile)
          z_axis:    1D Tensor of shape (num, ) with the depth coordinates.
          profile:   2D Tensor of shape (batch_size, num) giving the SLD at each depth.
          d_profile: 2D Tensor of shape (batch_size, num) giving d(SLD)/dz at each depth.
    """

    bs, n = thicknesses.shape
    assert roughnesses.shape == (bs, n + 1),  (
        f"Roughnesses must be (batch_size, num_layers+1). Found {roughnesses.shape} instead."
    )
    assert slds.shape == (bs, n + 1), (
        f"SLDs must be (batch_size, num_layers+1). Found {slds.shape} instead."
    )
    assert torch.all(thicknesses >= 0), "Negative thickness encountered."
    assert torch.all(roughnesses >= 0), "Negative roughness encountered."

    if ambient_sld is None:
        ambient_sld = torch.zeros((bs, 1), device=thicknesses.device)
    else:
        if ambient_sld.ndim == 1:
            ambient_sld = ambient_sld.unsqueeze(-1)
        ambient_sld = ambient_sld.expand(bs, 1)

    slds_all = torch.cat([ambient_sld, slds], dim=-1)  # new dimension: n+2
    d_rhos = torch.diff(slds_all, dim=-1)  # (bs, n+1)

    interfaces = torch.cat([
        torch.zeros((bs, 1), device=thicknesses.device),  # z=0 for ambient→layer1
        thicknesses
    ], dim=-1).cumsum(dim=-1)  # now shape => (bs, n+1)

    total_thickness = interfaces[..., -1].max()
    if z_axis is None:
        z_axis = torch.linspace(
            -padding_left * total_thickness,
            padding_right * total_thickness,
            num,
            device=thicknesses.device
        )  # shape => (num,)
    if z_axis.ndim == 1:
        z_axis = z_axis.unsqueeze(0)  # shape => (1, num)

    z_b = z_axis.repeat(bs, 1).unsqueeze(1)           # (bs, 1, num)
    interfaces_b = interfaces.unsqueeze(-1)           # (bs, n+1, 1)
    sigmas_b = (roughnesses * sqrt(2)).unsqueeze(-1)  # (bs, n+1, 1)
    d_rhos_b = d_rhos.unsqueeze(-1)                   # (bs, n+1, 1)

    profile = get_erf(z_b, interfaces_b, sigmas_b, d_rhos_b).sum(dim=1)    # (bs, num)
    if ambient_sld is not None:
        profile = profile + ambient_sld

    d_profile = get_gauss(z_b, interfaces_b, sigmas_b, d_rhos_b).sum(dim=1) # (bs, num)

    return z_axis.squeeze(0), profile, d_profile

def get_param_labels(
        num_layers: int, *,
        thickness_name: str = 'Thickness',
        roughness_name: str = 'Roughness',
        sld_name: str = 'SLD',
        imag_sld_name: str = 'SLD imag',
        substrate_name: str = 'sub',
        parameterization_type: str = 'standard',
        number_top_to_bottom: bool = False,
) -> List[str]:
    def pos(i):
        return i + 1 if number_top_to_bottom else num_layers - i
        
    thickness_labels = [f'{thickness_name} L{pos(i)}' for i in range(num_layers)]
    roughness_labels = [f'{roughness_name} L{pos(i)}' for i in range(num_layers)] + [f'{roughness_name} {substrate_name}']
    sld_labels = [f'{sld_name} L{pos(i)}' for i in range(num_layers)] + [f'{sld_name} {substrate_name}']

    all_labels = thickness_labels + roughness_labels + sld_labels

    if parameterization_type == 'absorption':
        imag_sld_labels = [f'{imag_sld_name} L{pos(i)}' for i in range(num_layers)] + [f'{imag_sld_name} {substrate_name}']
        all_labels = all_labels + imag_sld_labels
        
    return all_labels



from typing import Dict, Union
import warnings

from torch import Tensor
import torch

from reflectorch.data_generation.priors import PriorSampler, BasicParams
from reflectorch.data_generation.noise import QNoiseGenerator, IntensityNoiseGenerator
from reflectorch.data_generation.q_generator import QGenerator
from reflectorch.data_generation.scale_curves import CurvesScaler, LogAffineCurvesScaler
from reflectorch.data_generation.smearing import Smearing

BATCH_DATA_TYPE = Dict[str, Union[Tensor, BasicParams]]


class BasicDataset(object):
    """Reflectometry dataset. It generates the q positions, samples the thin film parameters from the prior, 
    simulates the reflectivity curves and applies noise to the curves.

    Args:
        q_generator (QGenerator): the momentum transfer (q) generator
        prior_sampler (PriorSampler): the prior sampler
        intensity_noise (IntensityNoiseGenerator, optional): the intensity noise generator. Defaults to None.
        q_noise (QNoiseGenerator, optional): the q noise generator. Defaults to None.
        curves_scaler (CurvesScaler, optional): the reflectivity curve scaler. Defaults to an instance of LogAffineCurvesScaler, 
                                                which scales the curves to the range [-1, 1], the minimum considered intensity being 1e-10.
        calc_denoised_curves (bool, optional): whether to add the curves without noise to the dictionary. Defaults to False.
        calc_nonsmeared_curves (bool, optional): whether to add the curves without smearing to the dictionary (only relevant when smearing is applied). Defaults to False.
        smearing (Smearing, optional): curve smearing generator. Defaults to None.
    """
    def __init__(self,
                 q_generator: QGenerator,
                 prior_sampler: PriorSampler,
                 intensity_noise: IntensityNoiseGenerator = None,
                 q_noise: QNoiseGenerator = None,
                 curves_scaler: CurvesScaler = None,
                 calc_denoised_curves: bool = False,
                 calc_nonsmeared_curves: bool = False,
                 smearing: Smearing = None,
                 ):
        self.q_generator = q_generator
        self.intensity_noise = intensity_noise
        self.q_noise = q_noise
        self.curves_scaler = curves_scaler or LogAffineCurvesScaler()
        self.prior_sampler = prior_sampler
        self.smearing = smearing
        self.calc_denoised_curves = calc_denoised_curves
        self.calc_nonsmeared_curves = calc_nonsmeared_curves

    def update_batch_data(self, batch_data: BATCH_DATA_TYPE) -> None:
        """implement in a subclass to edit batch_data dict inplace"""
        pass

    def _sample_from_prior(self, batch_size: int):
        params: BasicParams = self.prior_sampler.sample(batch_size)
        scaled_params: Tensor = self.prior_sampler.scale_params(params)
        return params, scaled_params

    def get_batch(self, batch_size: int) -> BATCH_DATA_TYPE:
        """get a batch of data as a dictionary with keys ``params``, ``scaled_params``, ``q_values``, ``curves``, ``scaled_noisy_curves``

        Args:
            batch_size (int): the batch size
        """
        batch_data = {}

        params, scaled_params = self._sample_from_prior(batch_size)

        batch_data['params'] = params
        batch_data['scaled_params'] = scaled_params

        q_values: Tensor = self.q_generator.get_batch(batch_size, batch_data)

        if self.q_noise:
            batch_data['original_q_values'] = q_values
            q_values = self.q_noise.apply(q_values, batch_data)

        batch_data['q_values'] = q_values

        refl_kwargs = {}

        curves, q_resolutions, nonsmeared_curves = self._calc_curves(q_values, params, refl_kwargs)

        if torch.is_tensor(q_resolutions):
            batch_data['q_resolutions'] = q_resolutions

        if torch.is_tensor(nonsmeared_curves):
            batch_data['nonsmeared_curves'] = nonsmeared_curves

        if self.calc_denoised_curves:
            batch_data['curves'] = curves

        noisy_curves = curves

        if self.intensity_noise:
            noisy_curves = self.intensity_noise(noisy_curves, batch_data)

        scaled_noisy_curves = self.curves_scaler.scale(noisy_curves)
        batch_data['scaled_noisy_curves'] = scaled_noisy_curves

        is_finite = torch.all(torch.isfinite(scaled_noisy_curves), -1)

        if not torch.all(is_finite).item():
            infinite_indices = ~is_finite
            to_recalculate = infinite_indices.sum().item()
            warnings.warn(f'Infinite number appeared in the curve simulation! Recalculate {to_recalculate} curves.')
            recalculated_batch_data = self.get_batch(to_recalculate)
            _insert_batch_data(batch_data, recalculated_batch_data, infinite_indices)

        is_finite = torch.all(torch.isfinite(batch_data['scaled_noisy_curves']), -1)
        assert torch.all(is_finite).item()

        self.update_batch_data(batch_data)

        return batch_data

    def _calc_curves(self, q_values: Tensor, params: BasicParams, refl_kwargs):
        nonsmeared_curves = None

        if self.smearing:
            if self.calc_nonsmeared_curves:
                nonsmeared_curves = params.reflectivity(q_values, **refl_kwargs)
            curves, q_resolutions = self.smearing.get_curves(q_values, params, refl_kwargs)
        else:
            curves = params.reflectivity(q_values, **refl_kwargs)
            q_resolutions = None

        curves = curves.to(q_values)
        return curves, q_resolutions, nonsmeared_curves


def _insert_batch_data(tgt_batch_data, add_batch_data, indices):
    for key in tuple(tgt_batch_data.keys()):
        value = tgt_batch_data[key]
        if isinstance(value, BasicParams) or len(value.shape) == 2:
            value[indices] = add_batch_data[key]
        else:
            warnings.warn(f'Ignore {key} while merging batch_data.')


if __name__ == '__main__':
    from reflectorch.data_generation.q_generator import ConstantQ
    from reflectorch.data_generation.priors import BasicPriorSampler, UniformSubPriorSampler
    from reflectorch.data_generation.noise import BasicExpIntensityNoise
    from reflectorch.data_generation.noise import BasicQNoiseGenerator
    from reflectorch.utils import to_np
    from time import perf_counter

    q_generator = ConstantQ((0, 0.2, 65), device='cpu')
    noise_gen = BasicExpIntensityNoise(
        relative_errors=(0.05, 0.2),
        # scale_range=(-1e-2, 1e-2),
        logdist=True,
        apply_shift=True,
    )
    q_noise_gen = BasicQNoiseGenerator(
        shift_std=5e-4,
        noise_std=(0, 1e-3),
    )
    prior_sampler = UniformSubPriorSampler(
        thickness_range=(0, 250),
        roughness_range=(0, 40),
        sld_range=(0, 60),
        num_layers=2,
        device=torch.device('cpu'),
        dtype=torch.float64,
        smaller_roughnesses=True,
        logdist=True,
        relative_min_bound_width=5e-4,
    )
    smearing = Smearing(
        sigma_range=(0.8e-3, 5e-3),
        gauss_num=31,
        share_smeared=0.5,
    )

    dataset = BasicDataset(
        q_generator,
        prior_sampler,
        noise_gen,
        q_noise=q_noise_gen,
        smearing=smearing
    )
    start = perf_counter()
    batch_data = dataset.get_batch(32)
    print(f'Total time = {(perf_counter() - start):.3f} sec ')
    print(batch_data['params'].roughnesses[:10])
    print(batch_data['scaled_noisy_curves'].min().item())

    scaled_noisy_curves = batch_data['scaled_noisy_curves']
    scaled_curves = dataset.curves_scaler.scale(
        batch_data['params'].reflectivity(q_generator.q)
    )

    try:
        import matplotlib.pyplot as plt

        for i in range(16):
            plt.plot(
                to_np(q_generator.q.squeeze().cpu().numpy()),
                to_np(scaled_curves[i])
            )
            plt.plot(
                to_np(q_generator.q.squeeze().cpu().numpy()),
                to_np(scaled_noisy_curves[i])
            )

            plt.show()
    except ImportError:
        pass



import warnings

import numpy as np
from scipy.optimize import minimize, curve_fit
import torch

from reflectorch.data_generation.priors.base import PriorSampler
from reflectorch.data_generation.reflectivity import abeles_np

__all__ = [
    "standard_refl_fit",
    "refl_fit",
    "fit_refl_curve",
    "restore_masked_params",
    "get_fit_with_growth",
]


def standard_restore_params(fitted_params) -> dict:
    num_layers = (fitted_params.size - 2) // 3

    return dict(
        thickness=fitted_params[:num_layers],
        roughness=fitted_params[num_layers:2 * num_layers + 1],
        sld=fitted_params[2 * num_layers + 1:],
    )


def mse_loss(curve1, curve2):
    return np.sum((curve1 - curve2) ** 2)

def standard_refl_fit(
        q: np.ndarray, curve: np.ndarray,
        init_params: np.ndarray,
        bounds: np.ndarray = None,
        refl_generator=abeles_np,
        restore_params_func=standard_restore_params,
        scale_curve_func=np.log10,
        **kwargs
):
    if bounds is not None:
        kwargs['bounds'] = bounds
        init_params = np.clip(init_params, *bounds)

    res = curve_fit(
        standard_get_scaled_curve_func(
            refl_generator=refl_generator,
            restore_params_func=restore_params_func,
            scale_curve_func=scale_curve_func,
        ),
        q, scale_curve_func(curve),
        p0=init_params, **kwargs
    )

    curve = refl_generator(q, **restore_params_func(res[0]))
    return res[0], curve

def refl_fit(
        q: np.ndarray, 
        curve: np.ndarray,
        init_params: np.ndarray,
        prior_sampler: PriorSampler,
        bounds: np.ndarray = None,
        error_bars: np.ndarray = None,
        scale_curve_func=np.log10,
        method: str = 'trf', #'lm', 'trf'
        polishing_max_nfev: int = None,
        reflectivity_kwargs: dict = None,
        **kwargs
):
    if bounds is not None:
        # introduce a small perturbation for fixed bounds
        epsilon = 1e-6
        adjusted_bounds = bounds.copy()

        for i in range(bounds.shape[1]): 
            if bounds[0, i] == bounds[1, i]:
                adjusted_bounds[0, i] -= epsilon
                adjusted_bounds[1, i] += epsilon

        init_params = np.clip(init_params, *adjusted_bounds)
        if method != 'lm':
            kwargs['bounds'] = adjusted_bounds

    reflectivity_kwargs = reflectivity_kwargs or {}
    for key, value in reflectivity_kwargs.items():
        if isinstance(value, float):
            reflectivity_kwargs[key] = torch.tensor([[value]], dtype=torch.float64)
        elif isinstance(value, np.ndarray):
            reflectivity_kwargs[key] = torch.tensor(value, dtype=torch.float32).unsqueeze(0)     

    curve = np.clip(curve, a_min=1e-12, a_max=None)

    if error_bars is not None and scale_curve_func == np.log10:
        error_bars = np.clip(error_bars, a_min=1e-20, a_max=None)
        scaled_error_bars = error_bars / (curve * np.log(10))
    else:
        scaled_error_bars = None  

    res = curve_fit(
        f=get_scaled_curve_func(
            scale_curve_func=scale_curve_func,
            prior_sampler=prior_sampler,
            reflectivity_kwargs=reflectivity_kwargs,
        ),
        xdata=q, 
        ydata=scale_curve_func(curve).reshape(-1),
        p0=init_params,
        sigma=scaled_error_bars,
        absolute_sigma=True,
        method=method,
        max_nfev=polishing_max_nfev,
        **kwargs
    )

    curve = prior_sampler.param_model.reflectivity(torch.tensor(q, dtype=torch.float64), 
                                                   torch.tensor(res[0], dtype=torch.float64).unsqueeze(0), 
                                                   **reflectivity_kwargs).squeeze().numpy()
    return res[0], curve


def get_fit_with_growth(
        q: np.ndarray, 
        curve: np.ndarray,
        init_params: np.ndarray,
        bounds: np.ndarray = None,
        init_d_change: float = 0.,
        max_d_change: float = 30.,
        scale_curve_func=np.log10,
        **kwargs
):
    init_params = np.array(list(init_params) + [init_d_change])
    if bounds is not None:
        bounds = np.concatenate([bounds, np.array([0, max_d_change])[..., None]], -1)

    params, curve = standard_refl_fit(
        q, 
        curve, 
        init_params, 
        bounds, 
        refl_generator=growth_reflectivity,
        restore_params_func=get_restore_params_with_growth_func(q_size=q.size, d_idx=0),
        scale_curve_func=scale_curve_func, 
        **kwargs
    )

    params[0] += params[-1] / 2
    return params, curve


def fit_refl_curve(q: np.ndarray, curve: np.ndarray,
                   init_params: np.ndarray,
                   bounds: np.ndarray = None,
                   refl_generator=abeles_np,
                   restore_params_func=standard_restore_params,
                   scale_curve_func=np.log10,
                   **kwargs
                   ) -> np.ndarray:
    fitting_func = get_fitting_func(
        q=q, curve=curve,
        refl_generator=refl_generator,
        restore_params_func=restore_params_func,
        scale_curve_func=scale_curve_func,
    )

    res = minimize(fitting_func, init_params, bounds=bounds, **kwargs)

    if not res.success:
        warnings.warn(f"Minimization did not converge.")
    return res.x

def standard_get_scaled_curve_func(
        refl_generator=abeles_np,
        restore_params_func=standard_restore_params,
        scale_curve_func=np.log10,
):
    def scaled_curve_func(q, *fitted_params):
        fitted_params = restore_params_func(np.asarray(fitted_params))
        fitted_curve = refl_generator(q, **fitted_params)
        scaled_curve = scale_curve_func(fitted_curve)
        return scaled_curve

    return scaled_curve_func

def get_scaled_curve_func(
        scale_curve_func=np.log10,
        prior_sampler: PriorSampler = None,
        reflectivity_kwargs: dict = None,
):  
    reflectivity_kwargs = reflectivity_kwargs or {}
    
    def scaled_curve_func(q, *fitted_params):
        q_tensor = torch.from_numpy(q).to(torch.float64)
        fitted_params_tensor = torch.tensor(fitted_params, dtype=torch.float64).unsqueeze(0)
        
        fitted_curve_tensor = prior_sampler.param_model.reflectivity(q_tensor, fitted_params_tensor, **reflectivity_kwargs)
        fitted_curve = fitted_curve_tensor.squeeze().numpy()
        
        scaled_curve = scale_curve_func(fitted_curve)

        return scaled_curve.reshape(-1)

    return scaled_curve_func


def get_fitting_func(
        q: np.ndarray,
        curve: np.ndarray,
        refl_generator=abeles_np,
        restore_params_func=standard_restore_params,
        scale_curve_func=np.log10,
        loss_func=mse_loss,
):
    scaled_curve = scale_curve_func(curve)

    def fitting_func(fitted_params):
        fitted_params = restore_params_func(fitted_params)
        fitted_curve = refl_generator(q, **fitted_params)
        loss = loss_func(scale_curve_func(fitted_curve), scaled_curve)
        return loss

    return fitting_func


def restore_masked_params(fixed_params, fixed_mask):
    def restore_params(fitted_params) -> dict:
        params = np.empty_like(fixed_mask).astype(fitted_params.dtype)
        params[fixed_mask] = fixed_params
        params[~fixed_mask] = fitted_params
        return standard_restore_params(params)

    return restore_params


def base_params2growth(base_params: dict, d_shift: np.ndarray, d_idx: int = 0) -> dict:
    d_init = base_params['thickness'][None]
    q_size = d_shift.size
    d = d_init.repeat(q_size, 0)
    d[:, d_idx] = d[:, d_idx] + d_shift

    roughness = np.broadcast_to(base_params['roughness'][None], (q_size, base_params['roughness'].size))
    sld = np.broadcast_to(base_params['sld'][None], (q_size, base_params['sld'].size))

    return {
        'thickness': d,
        'roughness': roughness,
        'sld': sld,
    }


def get_restore_params_with_growth_func(q_size: int, d_idx: int = 0):
    def restore_params_with_growth(fitted_params) -> dict:
        fitted_params, delta_d = fitted_params[:-1], fitted_params[-1]
        base_params = standard_restore_params(fitted_params)
        d_shift = np.linspace(0, delta_d, q_size)
        return base_params2growth(base_params, d_shift, d_idx)

    return restore_params_with_growth


def growth_reflectivity(q: np.ndarray, **kwargs):
    return abeles_np(q[..., None], **kwargs).flatten()



import os
import tempfile
import yaml
from huggingface_hub import hf_hub_download, list_repo_files

class HuggingfaceQueryMatcher:
    """Downloads the available configurations files to a temporary directory and provides functionality for filtering those configuration files matching user specified queries.
    
    Args:
        repo_id (str): The Hugging Face repository ID.
        config_dir (str): Directory within the repo where YAML files are stored.
    """
    def __init__(self, repo_id='valentinsingularity/reflectivity', config_dir='configs'):
        self.repo_id = repo_id
        self.config_dir = config_dir
        self.cache = {
            'parsed_configs': None,
            'temp_dir': None
        }
        self._renew_cache()

    def _renew_cache(self):
        temp_dir = tempfile.mkdtemp()
        print(f"Temporary directory created at: {temp_dir}")

        repo_files = list_repo_files(self.repo_id, repo_type='model')
        config_files = [file for file in repo_files if file.startswith(self.config_dir) and file.endswith('.yaml')]

        downloaded_files = []
        for file in config_files:
            file_path = hf_hub_download(repo_id=self.repo_id, filename=file, local_dir=temp_dir, repo_type='model')
            downloaded_files.append(file_path)
        
        parsed_configs = {}
        for file_path in downloaded_files:
            with open(file_path, 'r') as file:
                config_data = yaml.safe_load(file)
                file_name = os.path.basename(file_path)
                parsed_configs[file_name] = config_data
        
        self.cache['parsed_configs'] = parsed_configs
        self.cache['temp_dir'] = temp_dir

    def get_matching_configs(self, query):
        """retrieves configuration files that match the user specified query.
        
        Args:
            query (dict): Dictionary of key-value pairs to filter configurations, e.g. ``query = {'dset.prior_sampler.kwargs.max_num_layers': 3, 'dset.prior_sampler.kwargs.param_ranges.slds': [0., 100.]}``.
                          For keys containing the ``param_ranges`` subkey a configuration is selected if the value of the query (i.e. desired parameter range) 
                          is a subrange of the parameter range in the configuration, in all other cases the values must match exactly.
        
        Returns:
            list: List of file names that match the query.
        """
        
        filtered_configs = []
        
        for file_name, config_data in self.cache['parsed_configs'].items():
            if self.matches_query(config_data, query):
                filtered_configs.append(file_name)
        
        return filtered_configs
    
    def matches_query(self, config_data, query):
        for q_key, q_value in query.items():
            keys = q_key.split('.')
            value = self.deep_get(config_data, keys)
            if 'param_ranges' in keys:
                if q_value[0] < value[0] or q_value[1] > value[1]:
                    return False
            else: 
                if value != q_value:
                    return False
                
        return True

    def deep_get(self, d, keys):
        for key in keys:
            if isinstance(d, dict):
                d = d.get(key, None)

        return d


import pytest
from reflectorch import ConstantQ, VariableQ


@pytest.mark.parametrize("batch_size", [1, 128])
@pytest.mark.parametrize("q_range", [(0.01, 0.3, 128), (0.05, 0.1, 200)])
def test_constantq(batch_size, q_range):
    q_generator = ConstantQ(q=q_range, device='cpu')
    q_values = q_generator.get_batch(batch_size=batch_size)

    assert q_values.shape == (batch_size, q_range[-1])
    assert q_values.min().float() == q_range[0]
    assert q_values.max().float() == q_range[1]

@pytest.mark.parametrize("batch_size", [1, 128])
@pytest.mark.parametrize("q_min_range", [(0.005, 0.05), (0.01, 0.01)])
@pytest.mark.parametrize("q_max_range", [(0.15, 0.5), (0.3, 0.3)])
@pytest.mark.parametrize("n_q_range", [(64, 256), (128, 128)])
def test_variableq(batch_size, q_min_range, q_max_range, n_q_range):
    q_generator = VariableQ(q_min_range, q_max_range, n_q_range, device='cpu')
    q_values = q_generator.get_batch(batch_size=batch_size)

    assert n_q_range[0] <= q_values.shape[-1] <= n_q_range[1]
    assert q_values.min().float() >= q_min_range[0]
    assert q_values.max().float() <= q_max_range[1]


import pytest
from reflectorch import ConstantQ, BasicExpIntensityNoise, BasicQNoiseGenerator, LogAffineCurvesScaler, SubpriorParametricSampler, \
      Smearing, BasicDataset, ReflectivityDataLoader

@pytest.mark.parametrize("calc_denoised_curves", [False, True])
def test_dataloader(calc_denoised_curves):
    prior_sampler = SubpriorParametricSampler(
        param_ranges={
            'thicknesses': [5.0, 300.0],
            'roughnesses': [0.0, 30.0],
            'slds': [0.0, 50.0],
        },
        bound_width_ranges={
            'thicknesses': [1e-2, 300.0],
            'roughnesses': [1e-2, 30.0],
            'slds': [1e-2, 5.0],
        },
        model_name = 'standard_model',
        device='cpu',
        max_num_layers=3,
        constrained_roughness = True,
        max_thickness_share = 0.5,
        scale_params_by_ranges=False,
    )

    q_generator = ConstantQ(
    q = [0.02, 0.15, 128], 
    device = 'cpu')

    intensity_noise_generator = BasicExpIntensityNoise(
    relative_errors = [0.0, 0.2],
    abs_errors = 0.0,
    consistent_rel_err = True,
    logdist = False,
    apply_shift = True,
    shift_range = [-0.3, 0.3],
    apply_scaling = True,
    scale_range = [-0.02, 0.02],
    apply_background = True,
    background_range = [1.0e-10, 1.0e-8],
    )

    q_noise_generator = BasicQNoiseGenerator(
        shift_std = 1.0e-3,
        noise_std = [0., 1.0e-3],
    )

    smearing_generator = Smearing(
        sigma_range = (0.0001, 0.005), 
        constant_dq = True, 
        gauss_num = 31, 
        share_smeared = 0.2,
    )

    curves_scaler = LogAffineCurvesScaler(
        weight = 0.2,
        bias = 1.0,
        eps = 1.0e-10,
    )

    data_loader = ReflectivityDataLoader(
        q_generator=q_generator, 
        prior_sampler=prior_sampler,
        intensity_noise=intensity_noise_generator,
        q_noise=q_noise_generator,
        smearing=smearing_generator,
        curves_scaler=curves_scaler,
        calc_denoised_curves=calc_denoised_curves,
        )


from tqdm import trange

import torch
from torch import nn, Tensor

from reflectorch.data_generation import LogLikelihood, reflectivity, PriorSampler


class ReflGradientFit(object):
    """Directly optimizes the thin film parameters using a Pytorch optimizer

    Args:
        q (Tensor): the q positions
        exp_curve (Tensor): the experimental reflectivity curve
        prior_sampler (PriorSampler): the prior sampler
        params (Tensor): the initial thin film parameters
        fit_indices (Tensor): the indices of the thin film parameters which are to be fitted
        sigmas (Tensor, optional): error bars of the reflectivity curve, if not provided they are derived from ``rel_err`` and ``abs_err``. Defaults to None.
        optim_cls (Type[torch.optim.Optimizer], optional): the Pytorch optimizer class. Defaults to None.
        lr (float, optional): the learning rate. Defaults to 1e-2.
        rel_err (float, optional): the relative error in the reflectivity curve. Defaults to 0.1.
        abs_err (float, optional): the absolute error in the reflectivity curve. Defaults to 1e-7.
    """
    def __init__(self,
                 q: Tensor,
                 exp_curve: Tensor,
                 prior_sampler: PriorSampler,
                 params: Tensor,
                 fit_indices: Tensor,
                 sigmas: Tensor = None,
                 optim_cls=None,
                 lr: float = 1e-2,
                 rel_err: float = 0.1,
                 abs_err: float = 1e-7,
                 ):
        self.q = q

        if sigmas is None:
            sigmas = exp_curve * rel_err + abs_err

        self.likelihood = LogLikelihood(q, exp_curve, prior_sampler, sigmas)

        self.num_layers = params.shape[-1] // 3
        self.fit_indices = fit_indices
        self.init_params = params.clone()
        self.params_to_fit = nn.Parameter(self.init_params[fit_indices].clone())

        optim_cls = optim_cls or torch.optim.Adam
        self.optim = optim_cls([self.params_to_fit], lr)

        self.losses = []

    @property
    def params(self):
        params = self.init_params.clone()
        params[self.fit_indices] = self.params_to_fit
        return params

    def calc_log_likelihood(self):
        return self.likelihood.calc_log_likelihood(self.refl())

    def calc_log_prob_loss(self):
        return - self.calc_log_likelihood().mean()

    def refl(self):
        d, sigma, rho = torch.split(self.params, [self.num_layers, self.num_layers + 1, self.num_layers + 1], -1)
        return reflectivity(self.q, d, sigma, rho)

    def run(self, num_iterations: int = 500, disable_tqdm: bool = False):
        """Runs the optimization process

        Args:
            num_iterations (int, optional): number of iterations the optimization is run for. Defaults to 500.
            disable_tqdm (bool, optional): whether to disable the prograss bar. Defaults to False.
        """
        pbar = trange(num_iterations, disable=disable_tqdm)

        for _ in pbar:
            self.optim.zero_grad()
            loss = self.calc_log_prob_loss()
            loss.backward()
            self.optim.step()
            self.losses.append(loss.item())
            pbar.set_description(f'Loss = {loss.item():.2e}')

    def clear(self):
        self.losses.clear()



import numpy as np

from reflectorch.inference import standard_preprocessing, StandardPreprocessing


def test_preprocessing(raw_data_with_preprocessing_params):
    raw_data, preprocessing_params = raw_data_with_preprocessing_params
    res = standard_preprocessing(**raw_data, **preprocessing_params)

    assert "curve_interp" in res
    assert "curve" in res
    assert "q_values" in res
    assert "q_interp" in res

    assert res["curve_interp"].shape == res["q_interp"].shape
    assert res["curve"].shape == res["q_values"].shape
    assert len(res["curve_interp"].shape) == 1
    assert len(res["curve"].shape) == 1


def test_preprocessing_obj(raw_data_with_preprocessing_params):
    raw_data, preprocessing_params = raw_data_with_preprocessing_params
    func_res = standard_preprocessing(**raw_data, **preprocessing_params)
    st_p = StandardPreprocessing()

    obj_res = st_p.preprocess(**raw_data, **preprocessing_params)

    for key in func_res:
        assert np.all(func_res[key] == obj_res[key])

    st_p.set_parameters(**preprocessing_params)
    obj_res = st_p(**raw_data)

    for key in func_res:
        assert np.all(func_res[key] == obj_res[key])





import pytest
import torch
from reflectorch.models.encoders import ConvEncoder, FnoEncoder

@pytest.mark.parametrize("input_dim", [128, 140])
@pytest.mark.parametrize("in_channels", [1])
@pytest.mark.parametrize("hidden_channels", [(32, 64, 128, 256, 512), (16, 32, 64, 128, 256)])
@pytest.mark.parametrize("dim_embedding", [64, 80])
@pytest.mark.parametrize("dim_avpool", [1, 4])
@pytest.mark.parametrize("use_batch_norm", [True, False])
@pytest.mark.parametrize("activation", ['relu', 'gelu'])
def test_cnn(input_dim, in_channels, hidden_channels, dim_embedding, dim_avpool, use_batch_norm, activation):
    embedding_net = ConvEncoder(
        in_channels=in_channels, 
        hidden_channels=hidden_channels,
        dim_embedding=dim_embedding,
        dim_avpool=dim_avpool,
        use_batch_norm=use_batch_norm,
        activation=activation,
        )
    
    batch_size = 4
    input_tensor = torch.randn((batch_size, in_channels, input_dim))

    assert embedding_net(input_tensor).shape == (batch_size, dim_embedding)

@pytest.mark.parametrize("input_dim", [128, 140])
@pytest.mark.parametrize("in_channels", [1, 2, 3])
@pytest.mark.parametrize("dim_embedding", [64, 80])
@pytest.mark.parametrize("modes", [16, 32])
@pytest.mark.parametrize("fusion_self_attention", [False, True])
def test_fno(input_dim, in_channels, dim_embedding, modes, fusion_self_attention):
    embedding_net = FnoEncoder(
        in_channels = in_channels,
        dim_embedding = dim_embedding,
        modes = modes,
        width_fno = 128,
        n_fno_blocks = 2,
        activation = 'gelu',
        fusion_self_attention=fusion_self_attention,
    )

    batch_size = 4
    input_tensor = torch.randn((batch_size, in_channels, input_dim))

    assert embedding_net(input_tensor).shape == (batch_size, dim_embedding)


import torch
from torch import Tensor

from reflectorch.data_generation.priors.parametric_subpriors import BasicParams


class Smearing(object):
    """Class which applies resolution smearing to the reflectivity curves. 
    The intensity at a q point will be the average of the intensities of neighbouring q points, weighted by a gaussian profile.

    Args:
        sigma_range (tuple, optional): the range for sampling the resolutions. Defaults to (0.01, 0.1).
        constant_dq (bool, optional): if ``True`` the smearing is constant (the resolution is given by the constant dq at each point in the curve) 
                                    otherwise the smearing is linear (the resolution is given by the constant dq/q at each point in the curve). Defaults to True.
        gauss_num (int, optional): the number of interpolating gaussian profiles. Defaults to 31.
        share_smeared (float, optional): the share of curves in the batch for which the resolution smearing is applied. Defaults to 0.2.
    """
    def __init__(self,
                 sigma_range: tuple = (0.01, 0.1),
                 constant_dq: bool = False,
                 gauss_num: int = 31,
                 share_smeared: float = 0.2,
                 ):
        self.sigma_min, self.sigma_max = sigma_range
        self.sigma_delta = self.sigma_max - self.sigma_min
        self.constant_dq = constant_dq
        self.gauss_num = gauss_num
        self.share_smeared = share_smeared

    def __repr__(self):
        return f'Smearing(({self.sigma_min}, {self.sigma_max})'

    def generate_resolutions(self, batch_size: int, device=None, dtype=None):
        num_smeared = int(batch_size * self.share_smeared)
        if not num_smeared:
            return None, None
        dq = torch.rand(num_smeared, 1, device=device, dtype=dtype) * self.sigma_delta + self.sigma_min
        indices = torch.zeros(batch_size, device=device, dtype=torch.bool)
        indices[torch.randperm(batch_size, device=device)[:num_smeared]] = True
        return dq, indices
    
    def scale_resolutions(self, resolutions: Tensor) -> Tensor:
        """Scales the q-resolution values to [-1,1] range using the internal sigma range"""
        sigma_min = 0.0 if self.share_smeared != 1.0 else self.sigma_min
        return 2 * (resolutions - sigma_min) / (self.sigma_max - sigma_min) - 1

    def get_curves(self, q_values: Tensor, params: BasicParams, refl_kwargs:dict = None):
        refl_kwargs = refl_kwargs or {}

        dq, indices = self.generate_resolutions(params.batch_size, device=params.device, dtype=params.dtype)
        q_resolutions = torch.zeros(q_values.shape[0], 1, dtype=q_values.dtype, device=q_values.device)

        if dq is None:
            return params.reflectivity(q_values, **refl_kwargs), q_resolutions
        
        refl_kwargs_not_smeared = {}
        refl_kwargs_smeared = {}
        for key, value in refl_kwargs.items():
            if isinstance(value, torch.Tensor) and value.shape[0] == params.batch_size:
                refl_kwargs_not_smeared[key] = value[~indices]
                refl_kwargs_smeared[key] = value[indices]
            else:
                refl_kwargs_not_smeared[key] = value
                refl_kwargs_smeared[key] = value

         # Compute unsmeared reflectivity
        if (~indices).sum().item():
            if q_values.dim() == 2 and q_values.shape[0] > 1:
                q = q_values[~indices]
            else:
                q = q_values

            reflectivity_not_smeared = params[~indices].reflectivity(q, **refl_kwargs_not_smeared)
        else:
            reflectivity_not_smeared = None

        # Compute smeared reflectivity
        if indices.sum().item():
            if q_values.dim() == 2 and q_values.shape[0] > 1:
                q = q_values[indices]
            else:
                q = q_values

            reflectivity_smeared = params[indices].reflectivity(
                q, dq=dq, constant_dq=self.constant_dq, gauss_num=self.gauss_num, **refl_kwargs_smeared
            )
        else:
            reflectivity_smeared = None

        curves = torch.empty(params.batch_size, q_values.shape[-1], device=params.device, dtype=params.dtype)
        
        if (~indices).sum().item():
            curves[~indices] = reflectivity_not_smeared
        
        curves[indices] = reflectivity_smeared

        q_resolutions[indices] = dq

        return curves, q_resolutions


import numpy as np
from functools import reduce
from operator import or_

from reflectorch.inference.inference_model import EasyInferenceModel
from reflectorch import BasicParams

import refnx
from refnx.dataset import ReflectDataset, Data1D
from refnx.analysis import Transform, CurveFitter, Objective, Model, Parameter
from refnx.reflect import SLD, Slab, ReflectModel

def covert_reflectorch_prediction_to_refnx_structure(inference_model: EasyInferenceModel, pred_params_object: BasicParams,  prior_bounds: np.array):
    assert inference_model.trainer.loader.prior_sampler.param_model.__class__.__name__ == 'StandardModel'

    n_layers = inference_model.trainer.loader.prior_sampler.max_num_layers
    init_thicknesses = pred_params_object.thicknesses.squeeze().tolist()
    init_roughnesses = pred_params_object.roughnesses.squeeze().tolist()
    init_slds = pred_params_object.slds.squeeze().tolist()

    sld_objects = []

    for sld in init_slds:
        sld_objects.append(SLD(value=sld))

    layer_objects = [SLD(0)()]
    for i in range(n_layers):
        layer_objects.append(sld_objects[i](init_thicknesses[i], init_roughnesses[i]))

    layer_objects.append(sld_objects[-1](0, init_roughnesses[-1]))

    thickness_bounds = prior_bounds[:n_layers]
    roughness_bounds = prior_bounds[n_layers:2*n_layers+1]
    sld_bounds = prior_bounds[2*n_layers+1:]

    for i, layer in enumerate(layer_objects):
        if i == 0:
            print("Ambient (air)")
            print(80 * '-')
        elif i < n_layers+1:
            layer.thick.setp(bounds=thickness_bounds[i-1], vary=True)
            layer.rough.setp(bounds=roughness_bounds[i-1], vary=True)
            layer.sld.real.setp(bounds=sld_bounds[i-1], vary=True)

            print(f'Layer {i}')
            print(f'Thickness: value {layer.thick.value}, vary {layer.thick.vary}, bounds {layer.thick.bounds}')
            print(f'Roughness: value {layer.rough.value}, vary {layer.rough.vary}, bounds {layer.rough.bounds}')
            print(f'SLD: value {layer.sld.real.value}, vary {layer.sld.real.vary}, bounds {layer.sld.real.bounds}')
            print(80 * '-')
        else: #substrate
            layer.rough.setp(bounds=roughness_bounds[i-1], vary=True)
            layer.sld.real.setp(bounds=sld_bounds[i-1], vary=True)

            print(f'Substrate')
            print(f'Thickness: value {layer.thick.value}, vary {layer.thick.vary}, bounds {layer.thick.bounds}')
            print(f'Roughness: value {layer.rough.value}, vary {layer.rough.vary}, bounds {layer.rough.bounds}')
            print(f'SLD: value {layer.sld.real.value}, vary {layer.sld.real.vary}, bounds {layer.sld.real.bounds}')
    
    refnx_structure = reduce(or_, layer_objects)

    return refnx_structure


###Example usage:
# refnx_structure = covert_reflectorch_prediction_to_refnx_structure(inference_model, pred_params_object, prior_bounds)

# refnx_reflect_model = ReflectModel(refnx_structure, bkg=1e-10, dq=0.0)
# refnx_reflect_model.scale.setp(bounds=(0.8, 1.2), vary=True)
# refnx_reflect_model.q_offset.setp(bounds=(-0.01, 0.01), vary=True)
# refnx_reflect_model.bkg.setp(bounds=(1e-10, 1e-8), vary=True)


# data = Data1D(data=(q_model, exp_curve_interp))

# refnx_objective = Objective(refnx_reflect_model, data, transform=Transform("logY"))
# fitter = CurveFitter(refnx_objective)
# fitter.fit('least_squares')


import pytest
import torch
from reflectorch import SubpriorParametricSampler

@pytest.mark.parametrize("batch_size", [1, 128])
@pytest.mark.parametrize("max_num_layers", [1, 2, 3, 4, 5, 10])
@pytest.mark.parametrize("constrained_roughness", [True, False])
def test_param_sampler_standard(batch_size, max_num_layers, constrained_roughness):
    param_sampler = SubpriorParametricSampler(
        param_ranges={
            'thicknesses': [10.0, 500.0],
            'roughnesses': [0.0, 60.0],
            'slds': [-20.0, 100.0],
        },
        bound_width_ranges={
            'thicknesses': [1e-2, 500.0],
            'roughnesses': [1e-2, 60.0],
            'slds': [1e-2, 10.0],
        },
        model_name = 'standard_model',
        device='cpu',
        max_num_layers=max_num_layers,
        constrained_roughness = constrained_roughness,
        max_thickness_share = 0.5,
        scale_params_by_ranges=False,
    )

    sampled_params = param_sampler.sample(batch_size=batch_size)
    assert sampled_params.parameters.shape == (batch_size, 3*max_num_layers+2)
    assert torch.all(sampled_params.min_bounds <= sampled_params.parameters ) and torch.all(sampled_params.parameters <= sampled_params.max_bounds)


@pytest.mark.parametrize("batch_size", [16])
@pytest.mark.parametrize("max_num_layers", [1, 3, 5])
@pytest.mark.parametrize("constrained_roughness", [True, False])
@pytest.mark.parametrize("constrained_isld", [True, False])
def test_param_sampler_absorption(batch_size, max_num_layers, constrained_roughness, constrained_isld):
    param_sampler = SubpriorParametricSampler(
        param_ranges={
            'thicknesses': [10.0, 500.0],
            'roughnesses': [0.0, 60.0],
            'slds': [-20.0, 100.0],
            'islds': [0.0, 10.0],
        },
        bound_width_ranges={
            'thicknesses': [1e-2, 500.0],
            'roughnesses': [1e-2, 60.0],
            'slds': [1e-3, 20.0],
            'islds': [1e-2, 10.0],
        },
        model_name = 'model_with_absorption',
        device='cpu',
        max_num_layers=max_num_layers,
        constrained_roughness = constrained_roughness,
        max_thickness_share = 0.5,
        constrained_isld = constrained_isld,
        max_sld_share = 0.2,
        scale_params_by_ranges=False,
    )

    sampled_params = param_sampler.sample(batch_size=batch_size)
    assert sampled_params.parameters.shape == (batch_size, 4*max_num_layers+3)
    assert torch.all(sampled_params.min_bounds <= sampled_params.parameters ) and torch.all(sampled_params.parameters <= sampled_params.max_bounds)


# -*- coding: utf-8 -*-
import math
from functools import reduce

import torch
from torch import Tensor


def abeles(
        q: Tensor,
        thickness: Tensor,
        roughness: Tensor,
        sld: Tensor,
):
    """Simulates reflectivity curves for SLD profiles with box model parameterization using the Abeles matrix method

    Args:
        q (Tensor): tensor of momentum transfer (q) values with shape [batch_size, n_points] or [n_points]
        thickness (Tensor): tensor containing the layer thicknesses (ordered from top to bottom) with shape [batch_size, n_layers]
        roughness (Tensor): tensor containing the interlayer roughnesses (ordered from top to bottom) with shape [batch_size, n_layers + 1]
        sld (Tensor): tensor containing the layer SLDs (real or complex; ordered from top to bottom). The tensor shape should be one of the following:
            - [batch_size, n_layers + 1]: in this case, the ambient SLD is not included but assumed to be 0
            - [batch_size, n_layers + 2]: this shape includes the ambient SLD as the first element in the tensor

    Returns:
        Tensor: tensor containing the simulated reflectivity curves with shape [batch_size, n_points]
    """
    c_dtype = torch.complex128 if q.dtype is torch.float64 else torch.complex64

    batch_size, num_layers = thickness.shape

    if sld.shape[-1] == num_layers + 1:
        # add zero ambient sld
        sld = torch.cat([torch.zeros(batch_size, 1).to(sld), sld], -1)
    if sld.shape[-1] != num_layers + 2:
        raise ValueError(
            "Number of SLD values does not equal to num_layers + 2 (substrate + ambient)."
        )

    sld = sld[:, None]

    # add zero thickness for ambient layer:
    thickness = torch.cat([torch.zeros(batch_size, 1).to(thickness), thickness], -1)[
        :, None
    ]

    roughness = roughness[:, None] ** 2

    sld = (sld - sld[..., :1]) * 1e-6 + 1e-36j

    k_z0 = (q / 2).to(c_dtype)

    if k_z0.dim() == 1:
        k_z0.unsqueeze_(0)

    if k_z0.dim() == 2:
        k_z0.unsqueeze_(-1)

    k_n = torch.sqrt(k_z0**2 - 4 * math.pi * sld)

    # k_n.shape - (batch, q, layers)

    k_n, k_np1 = k_n[..., :-1], k_n[..., 1:]

    beta = 1j * thickness * k_n

    exp_beta = torch.exp(beta)
    exp_m_beta = torch.exp(-beta)

    rn = (k_n - k_np1) / (k_n + k_np1) * torch.exp(-2 * k_n * k_np1 * roughness)

    c_matrices = torch.stack(
        [
            torch.stack([exp_beta, rn * exp_m_beta], -1),
            torch.stack([rn * exp_beta, exp_m_beta], -1),
        ],
        -1,
    )

    c_matrices = [c.squeeze(-3) for c in c_matrices.split(1, -3)]

    m = reduce(torch.matmul, c_matrices)

    r = (m[..., 1, 0] / m[..., 0, 0]).abs() ** 2
    r = torch.clamp_max_(r, 1.0)

    return r


# @torch.jit.script # commented so far due to complex numbers issue
def abeles_compiled(
        q: Tensor,
        thickness: Tensor,
        roughness: Tensor,
        sld: Tensor,
):
    return abeles(q, thickness, roughness, sld)



from reflectorch.data_generation import *
from reflectorch.ml import *
from reflectorch.models import *
from reflectorch.utils import *
from reflectorch.paths import *
from reflectorch.runs import *
from reflectorch.inference import *

from reflectorch.data_generation import __all__ as all_data_generation
from reflectorch.ml import __all__ as all_ml
from reflectorch.models import __all__ as all_models
from reflectorch.utils import __all__ as all_utils
from reflectorch.paths import __all__ as all_paths
from reflectorch.runs import __all__ as all_runs
from reflectorch.inference import __all__ as all_inference

__all__ = all_data_generation + all_ml + all_models + all_utils + all_paths + all_runs + all_inference



from IPython.display import clear_output

from ...ml import TrainerCallback, Trainer

from ..matplotlib import plot_losses


class JPlotLoss(TrainerCallback):
    """Callback for plotting the loss in a Jupyter notebook
    """
    def __init__(self, frequency: int, log: bool = True, clear: bool = True, **kwargs):
        """

        Args:
            frequency (int): plotting frequency
            log (bool, optional): if True, the plot is on a logarithmic scale. Defaults to True.
            clear (bool, optional):
        """
        self.frequency = frequency
        self.log = log
        self.kwargs = kwargs
        self.clear = clear

    def end_batch(self, trainer: Trainer, batch_num: int) -> None:
        if not batch_num % self.frequency:
            if self.clear:
                clear_output(wait=True)

            plot_losses(
                trainer.losses,
                log=self.log,
                best_epoch=trainer.callback_params.get('saved_iteration', None),
                **self.kwargs
            )



# -*- coding: utf-8 -*-

from math import pi

import torch
from torch import Tensor


def abeles_memory_eff(
        q: Tensor,
        thickness: Tensor,
        roughness: Tensor,
        sld: Tensor,
):
    """Simulates reflectivity curves for SLD profiles with box model parameterization using a memory-efficient implementation the Abeles matrix method. 
    It is computationally slower compared to the implementation in the 'abeles' function.

    Args:
        q (Tensor): tensor of momentum transfer (q) values with shape [batch_size, n_points] or [n_points]
        thickness (Tensor): tensor containing the layer thicknesses (ordered from top to bottom) with shape [batch_size, n_layers]
        roughness (Tensor): tensor containing the interlayer roughnesses (ordered from top to bottom) with shape [batch_size, n_layers + 1]
        sld (Tensor): tensor containing the layer SLDs (real or complex; ordered from top to bottom) with shape [batch_size, n_layers + 1]. 
                    It includes the substrate but excludes the ambient medium which is assumed to have an SLD of 0.

    Returns:
        Tensor: tensor containing the simulated reflectivity curves with shape [batch_size, n_points]
    """
    c_dtype = torch.complex128 if q.dtype is torch.float64 else torch.complex64

    batch_size, num_layers = thickness.shape

    sld = sld * 1e-6 + 1e-30j

    num_interfaces = num_layers + 1

    k_z0 = (q / 2).to(c_dtype)

    if len(k_z0.shape) == 1:
        k_z0.unsqueeze_(0)

    thickness_prev_layer = 1.  # ambient

    for interface_num in range(num_interfaces):

        prev_layer_idx = interface_num - 1
        next_layer_idx = interface_num

        if interface_num == 0:
            k_z_previous_layer = _get_relative_k_z(k_z0, torch.zeros(batch_size, 1).to(sld))
        else:
            thickness_prev_layer = thickness[:, prev_layer_idx].unsqueeze(1)
            k_z_previous_layer = _get_relative_k_z(k_z0, sld[:, prev_layer_idx].unsqueeze(1))

        k_z_next_layer = _get_relative_k_z(k_z0, sld[:, next_layer_idx].unsqueeze(1))  # (batch_num, q_num)

        reflection_matrix = _make_reflection_matrix(
            k_z_previous_layer, k_z_next_layer, roughness[:, interface_num].unsqueeze(1)
        )

        if interface_num == 0:
            total_reflectivity_matrix = reflection_matrix
        else:
            translation_matrix = _make_translation_matrix(k_z_previous_layer, thickness_prev_layer)

            total_reflectivity_matrix = torch.einsum(
                'bnmr, bmlr, bljr -> bnjr', total_reflectivity_matrix, translation_matrix, reflection_matrix
            )

    r = total_reflectivity_matrix[:, 0, 1] / total_reflectivity_matrix[:, 1, 1]

    reflectivity = torch.clamp_max_(torch.abs(r) ** 2, 1.).flatten(1)

    return reflectivity


def _get_relative_k_z(k_z0, scattering_length_density):
    return torch.sqrt(k_z0 ** 2 - 4 * pi * scattering_length_density)


def _make_reflection_matrix(k_z_previous_layer, k_z_next_layer, interface_roughness):
    p = _safe_div((k_z_previous_layer + k_z_next_layer), (2 * k_z_previous_layer)) * \
        torch.exp(-(k_z_previous_layer - k_z_next_layer) ** 2 * 0.5 * interface_roughness ** 2)

    m = _safe_div((k_z_previous_layer - k_z_next_layer), (2 * k_z_previous_layer)) * \
        torch.exp(-(k_z_previous_layer + k_z_next_layer) ** 2 * 0.5 * interface_roughness ** 2)

    return _stack_mtx(p, m, m, p)


def _stack_mtx(a11, a12, a21, a22):
    return torch.stack([
        torch.stack([a11, a12], dim=1),
        torch.stack([a21, a22], dim=1),
    ], dim=1)


def _make_translation_matrix(k_z, thickness):
    return _stack_mtx(
        torch.exp(-1j * k_z * thickness), torch.zeros_like(k_z),
        torch.zeros_like(k_z), torch.exp(1j * k_z * thickness)
    )


def _safe_div(numerator, denominator):
    return torch.where(denominator == 0, numerator, torch.divide(numerator, denominator))



import torch
from torch import nn
from torch.nn.functional import relu

class Rowdy(nn.Module):
    """adaptive activation function"""
    def __init__(self, K=9):
        super().__init__()
        self.K = K
        self.alpha = nn.Parameter(torch.cat((torch.ones(1), torch.zeros(K-1))))  
        self.alpha.requiresGrad = True
        self.omega = nn.Parameter(torch.ones(K))  
        self.omega.requiresGrad = True

    def forward(self, x):
        rowdy = self.alpha[0]*relu(self.omega[0]*x)
        for k in range(1, self.K):
            rowdy += self.alpha[k]*torch.sin(self.omega[k]*k*x)
        return rowdy
    

ACTIVATIONS = {
    'relu': nn.ReLU,
    'lrelu': nn.LeakyReLU,
    'gelu': nn.GELU,
    'selu': nn.SELU,
    'elu': nn.ELU,
    'sigmoid': nn.Sigmoid,    
    'tanh': nn.Tanh,
    'silu': nn.SiLU,
    'mish': nn.Mish,
    'rowdy': Rowdy,
}


def activation_by_name(name):
    """returns an activation function module corresponding to its name

    Args:
        name (str): string denoting the activation function ('relu', 'lrelu', 'gelu', 'selu', 'elu', 'sigmoid', 'silu', 'mish', 'rowdy')

    Returns:
        nn.Module: Pytorch activation function module
    """
    if not isinstance(name, str):
        return name
    try:
        return ACTIVATIONS[name.lower()]
    except KeyError:
        raise KeyError(f'Unknown activation function {name}')



import torch
from torch import nn
from torch.nn import functional as F
from torch.nn import init

from reflectorch.models.activations import activation_by_name

class ResidualMLP(nn.Module):
    """Multilayer perceptron with residual blocks (BN-Act-Linear-BN-Act-Linear)"""

    def __init__(
            self,
            dim_in: int,
            dim_out: int,
            dim_condition: int = 0,
            layer_width: int = 512,
            num_blocks: int = 4,
            repeats_per_block: int = 2,
            activation: str = 'relu',
            use_batch_norm: bool = True,
            use_layer_norm: bool = False,
            dropout_rate: float = 0.0,
            residual: bool = True,
            adaptive_activation: bool = False,
            conditioning: str = 'glu',
            concat_condition_first_layer: bool = True,
            film_with_tanh: bool = False,
    ):
        super().__init__()

        self.concat_condition_first_layer = concat_condition_first_layer

        dim_first_layer = dim_in + dim_condition if concat_condition_first_layer else dim_in
        self.first_layer = nn.Linear(dim_first_layer, layer_width)

        self.blocks = nn.ModuleList(
            [
                ResidualBlock(
                    layer_width=layer_width,
                    dim_condition=dim_condition,
                    repeats_per_block=repeats_per_block,
                    activation=activation,
                    use_batch_norm=use_batch_norm,
                    use_layer_norm=use_layer_norm,
                    dropout_rate=dropout_rate,
                    residual=residual,
                    adaptive_activation=adaptive_activation,
                    conditioning = conditioning,
                    film_with_tanh = film_with_tanh,
                )
                for _ in range(num_blocks)
            ]
        )

        self.last_layer = nn.Linear(layer_width, dim_out)

    def forward(self, x, condition=None):
        if self.concat_condition_first_layer and condition is not None:
            x = self.first_layer(torch.cat([x, condition], dim=-1))
        else:
            x = self.first_layer(x)

        for block in self.blocks:
            x = block(x, condition=condition)

        x = self.last_layer(x)

        return x


class ResidualBlock(nn.Module):
    """Residual block (BN-Act-Linear-BN-Act-Linear)"""

    def __init__(
            self,
            layer_width: int,
            dim_condition: int = 0,
            repeats_per_block: int = 2,
            activation: str = 'relu',
            use_batch_norm: bool = False,
            use_layer_norm: bool = False,
            dropout_rate: float = 0.0,
            residual: bool = True,
            adaptive_activation: bool = False,
            conditioning: str = 'glu',
            film_with_tanh: bool = False,
    ):
        super().__init__()
         
        self.residual = residual
        self.repeats_per_block = repeats_per_block
        self.use_batch_norm = use_batch_norm
        self.use_layer_norm = use_layer_norm
        self.dropout_rate = dropout_rate
        self.adaptive_activation = adaptive_activation
        self.conditioning = conditioning
        self.film_with_tanh = film_with_tanh

        if not adaptive_activation:
            self.activation = activation_by_name(activation)()
        else:
            self.activation_layers = nn.ModuleList(
                [activation_by_name(activation)() for _ in range(repeats_per_block)]
            )

        if use_batch_norm:
            self.batch_norm_layers = nn.ModuleList(
                [nn.BatchNorm1d(layer_width, eps=1e-3) for _ in range(repeats_per_block)]
             )
        elif use_layer_norm:
            self.layer_norm_layers = nn.ModuleList(
                [nn.LayerNorm(layer_width) for _ in range(repeats_per_block)]
             )

        if dim_condition:
            if conditioning == 'glu':
                self.condition_layer = nn.Linear(dim_condition, layer_width)
            elif conditioning == 'film':
                self.condition_layer = nn.Linear(dim_condition, 2*layer_width)

        self.linear_layers = nn.ModuleList(
            [nn.Linear(layer_width, layer_width) for _ in range(repeats_per_block)]
        )

        if self.dropout_rate > 0:
            self.dropout = nn.Dropout(p=dropout_rate)

    def forward(self, x, condition=None):
        x0 = x
         
        for i in range(self.repeats_per_block):
            if self.use_batch_norm:
                x = self.batch_norm_layers[i](x)
            elif self.use_layer_norm:
                x = self.layer_norm_layers[i](x)

            if not self.adaptive_activation:
                x = self.activation(x)
            else:
                x = self.activation_layers[i](x)

            if self.dropout_rate > 0 and i == self.repeats_per_block - 1:
                x = self.dropout(x)

            x = self.linear_layers[i](x)
        
        if condition is not None:
            if self.conditioning == 'glu':
                x = F.glu(torch.cat((x, self.condition_layer(condition)), dim=-1), dim=-1)
            elif self.conditioning == 'film':
                gamma, beta = torch.chunk(self.condition_layer(condition), chunks=2, dim=-1)
                if self.film_with_tanh:
                    tanh = nn.Tanh()
                    gamma, beta = tanh(gamma), tanh(beta)
                x = x * gamma + beta

        return x0 + x if self.residual else x


from typing import Tuple, Dict

import numpy as np
import torch
from torch import Tensor

from reflectorch.data_generation.priors.base import PriorSampler
from reflectorch.data_generation.priors.params import Params
from reflectorch.data_generation.priors.no_constraints import (
    DEFAULT_DEVICE,
    DEFAULT_DTYPE,
)

from reflectorch.data_generation.priors.multilayer_models import MULTILAYER_MODELS, MultilayerModel
from reflectorch.utils import to_t


class MultilayerStructureParams(Params):
    pass


class SimpleMultilayerSampler(PriorSampler):
    PARAM_CLS = MultilayerStructureParams

    def __init__(self,
                 params: Dict[str, Tuple[float, float]],
                 model_name: str,
                 device: torch.device = DEFAULT_DEVICE,
                 dtype: torch.dtype = DEFAULT_DTYPE,
                 max_num_layers: int = 50,
                 ):
        self.multilayer_model: MultilayerModel = MULTILAYER_MODELS[model_name](max_num_layers)
        self.device = device
        self.dtype = dtype
        self.num_layers = max_num_layers
        ordered_bounds = [params[k] for k in self.multilayer_model.PARAMETER_NAMES]
        self._np_bounds = np.array(ordered_bounds).T
        self.min_bounds, self.max_bounds = torch.tensor(ordered_bounds, device=device, dtype=dtype).T[:, None]
        self._param_dim = len(params)

    @property
    def max_num_layers(self) -> int:
        return self.num_layers

    @property
    def param_dim(self) -> int:
        return self._param_dim

    def sample(self, batch_size: int) -> MultilayerStructureParams:
        return self.optimized_sample(batch_size)[0]

    def optimized_sample(self, batch_size: int) -> Tuple[MultilayerStructureParams, Tensor]:
        scaled_params = torch.rand(
            batch_size,
            self.min_bounds.shape[-1],
            device=self.min_bounds.device,
            dtype=self.min_bounds.dtype,
        )

        targets = self.restore_params(scaled_params)

        return targets, scaled_params

    def get_np_bounds(self):
        return np.array(self._np_bounds)

    def restore_np_params(self, params: np.ndarray):
        p = self.multilayer_model.to_standard_params(
            torch.atleast_2d(to_t(params))
        )

        return {
            'thickness': p['thicknesses'].squeeze().cpu().numpy(),
            'roughness': p['roughnesses'].squeeze().cpu().numpy(),
            'sld': p['slds'].squeeze().cpu().numpy()
        }

    def restore_params2parametrized(self, scaled_params: Tensor) -> Tensor:
        return scaled_params * (self.max_bounds - self.min_bounds) + self.min_bounds

    def restore_params(self, scaled_params: Tensor) -> MultilayerStructureParams:
        return self.to_standard_params(self.restore_params2parametrized(scaled_params))

    def to_standard_params(self, params: Tensor) -> MultilayerStructureParams:
        return MultilayerStructureParams(**self.multilayer_model.to_standard_params(params))

    def scale_params(self, params: Params) -> Tensor:
        raise NotImplementedError

    def log_prob(self, params: Params) -> Tensor:
        raise NotImplementedError

    def get_indices_within_domain(self, params: Params) -> Tensor:
        raise NotImplementedError

    def get_indices_within_bounds(self, params: Params) -> Tensor:
        raise NotImplementedError

    def filter_params(self, params: Params) -> Params:
        indices = self.get_indices_within_domain(params)
        return params[indices]

    def clamp_params(self, params: Params) -> Params:
        raise NotImplementedError



from typing import Tuple

import torch
from torch import Tensor
import numpy as np

from reflectorch.inference.inference_model import (
    InferenceModel,
)
from reflectorch.data_generation.reflectivity import kinematical_approximation_np, abeles_np

from reflectorch.data_generation.priors import (
    MultilayerStructureParams,
    SimpleMultilayerSampler,
)
from reflectorch.inference.record_time import print_time
from reflectorch.inference.scipy_fitter import standard_refl_fit
from reflectorch.inference.multilayer_fitter import MultilayerFit


class MultilayerInferenceModel(InferenceModel):
    def predict(self,
                intensity: np.ndarray,
                scattering_angle: np.ndarray,
                attenuation: np.ndarray,
                priors: np.ndarray = None,
                preprocessing_parameters: dict = None,
                polish: bool = True,
                use_raw_q: bool = False,
                **kwargs
                ) -> dict:

        with print_time("everything"):
            with print_time("preprocess"):
                preprocessed_dict = self.preprocess(
                    intensity, scattering_angle, attenuation, **(preprocessing_parameters or {})
                )

            preprocessed_curve = preprocessed_dict["curve_interp"]

            raw_curve, raw_q = preprocessed_dict["curve"], preprocessed_dict["q_values"]

            with print_time("predict_from_preprocessed_curve"):
                preprocessed_dict.update(self.predict_from_preprocessed_curve(
                    preprocessed_curve, priors,
                    raw_curve=(raw_curve if use_raw_q else None),
                    raw_q=raw_q,
                    polish=polish,
                    use_raw_q=use_raw_q,
                    **kwargs
                ))

            return preprocessed_dict

    def predict_from_preprocessed_curve(self,
                                        curve: np.ndarray,
                                        priors: np.ndarray = None, *,  # ignore the priors so far
                                        polish: bool = True,
                                        raw_curve: np.ndarray = None,
                                        raw_q: np.ndarray = None,
                                        clip_prediction: bool = True,
                                        use_raw_q: bool = False,
                                        use_sampler: bool = False,
                                        fitted_time_limit: float = 3.,
                                        sampler_rel_bounds: float = 0.3,
                                        polish_with_abeles: bool = False,
                                        **kwargs
                                        ) -> dict:

        scaled_curve = self._scale_curve(curve)

        predicted_params, parametrized = self._simple_prediction(scaled_curve)

        if use_sampler:
            parametrized: Tensor = self._sampler_solution(
                curve, parametrized,
                time_limit=fitted_time_limit,
                rel_bounds=sampler_rel_bounds,
            )

        init_raw_q = raw_q

        if raw_curve is None:
            raw_curve = curve
            raw_q = self.q.squeeze().cpu().numpy()
            raw_q_t = self.q
        else:
            raw_q_t = torch.from_numpy(raw_q).to(self.q)

        # if q_ratio != 1.:
        #     predicted_params.scale_with_q(q_ratio)
        #     raw_q = raw_q * q_ratio
        #     raw_q_t = raw_q_t * q_ratio

        prediction_dict = {
            "params": parametrized.squeeze().cpu().numpy(),
            "param_names": list(self._prior_sampler.multilayer_model.PARAMETER_NAMES),
            "curve_predicted": predicted_params.reflectivity(raw_q_t).squeeze().cpu().numpy()
        }

        # sld_x_axis, sld_profile, _ = get_density_profiles(
        #     predicted_params.thicknesses, predicted_params.roughnesses, predicted_params.slds, num=1024,
        # )
        #
        # prediction_dict['sld_profile'] = sld_profile.squeeze().cpu().numpy()
        # prediction_dict['sld_x_axis'] = sld_x_axis.squeeze().cpu().numpy()

        if polish:
            prediction_dict.update(self._polish_prediction(
                raw_q, raw_curve, parametrized, q_values=init_raw_q, use_kinematical=True
            ))
        if polish_with_abeles:
            prediction_dict.update(self._polish_prediction(
                raw_q, raw_curve, parametrized, q_values=init_raw_q, use_kinematical=False
            ))

        return prediction_dict

    def _simple_prediction(self, scaled_curve) -> Tuple[MultilayerStructureParams, Tensor]:
        with torch.no_grad():
            self.trainer.model.eval()
            scaled_params = self.trainer.model(scaled_curve)

        predicted_params, parametrized = self._restore_predicted_params(scaled_params)
        return predicted_params, parametrized

    def _restore_predicted_params(self, scaled_params: Tensor) -> Tuple[MultilayerStructureParams, Tensor]:
        parametrized = self._prior_sampler.restore_params2parametrized(scaled_params)
        predicted_params: MultilayerStructureParams = self._prior_sampler.restore_params(scaled_params)
        return predicted_params, parametrized

    @print_time
    def _sampler_solution(
            self,
            curve: Tensor or np.ndarray,
            predicted_params: Tensor,
            batch_size: int = 2 ** 13,
            time_limit: float = 3.,
            rel_bounds: float = 0.3,
    ) -> Tensor:

        fit_obj = MultilayerFit.from_prediction(
            predicted_params, self._prior_sampler, self.q, torch.as_tensor(curve).to(self.q),
            batch_size=batch_size, rel_bounds=rel_bounds,
        )

        fit_obj.run_fixed_time(time_limit)

        best_params = fit_obj.get_best_solution()

        return best_params

    @property
    def _prior_sampler(self) -> SimpleMultilayerSampler:
        return self.trainer.loader.prior_sampler

    @print_time
    def _polish_prediction(self,
                           q: np.ndarray,
                           curve: np.ndarray,
                           predicted_params: Tensor,
                           q_values: np.ndarray,
                           use_kinematical: bool = True,
                           ) -> dict:

        params = predicted_params.squeeze().cpu().numpy()
        polished_params_dict = {}

        if use_kinematical:
            refl_generator = kinematical_approximation_np
        else:
            refl_generator = abeles_np

        try:
            polished_params_arr, curve_polished = standard_refl_fit(
                q, curve, params, restore_params_func=self._prior_sampler.restore_np_params,
                refl_generator=refl_generator,
                bounds=self._prior_sampler.get_np_bounds(),
            )
            params = self._prior_sampler.restore_np_params(polished_params_arr)
            if q_values is None:
                q_values = q
            curve_polished = abeles_np(q_values, **params)

        except Exception as err:
            self.log.exception(err)
            polished_params_arr = params
            curve_polished = np.zeros_like(q)

        polished_params_dict['params_polished'] = polished_params_arr
        polished_params_dict['curve_polished'] = curve_polished

        return polished_params_dict



import numpy as np

from reflectorch.utils import angle_to_q


def cut_curve(q: np.ndarray, curve: np.ndarray, max_q: float, max_angle: float, wavelength: float):
    """Cuts an experimental reflectivity curve at a maximum q position

    Args:
        q (np.ndarray): the array of q points
        curve (np.ndarray): the experimental reflectivity curve
        max_q (float): the maximum q value at which the curve is cut
        max_angle (float): the maximum scattering angle at which the curve is cut; only used if max_q is not provided
        wavelength (float): the wavelength of the beam

    Returns:
        tuple: the q array after cutting, the reflectivity curve after cutting, and the ratio between the maximum q after cutting and before cutting
    """
    if max_angle is None and max_q is None:
        q_ratio = 1.
    else:
        if max_q is None:
            max_q = angle_to_q(max_angle, wavelength)

        q_ratio = max_q / q.max()

        if q_ratio < 1.:
            idx = np.argmax(q > max_q)
            q = q[:idx] / q_ratio
            curve = curve[:idx]
    return q, curve, q_ratio



from reflectorch.runs.train import (
    run_train,
    run_train_on_cluster,
    run_test_config,
)

from reflectorch.runs.utils import (
    train_from_config,
    get_trainer_from_config,
    get_paths_from_config,
    get_callbacks_from_config,
    get_trainer_by_name,
    get_callbacks_by_name,
    convert_pt_to_safetensors,
)

from reflectorch.runs.config import load_config

__all__ = [
    'run_train',
    'run_train_on_cluster',
    'train_from_config',
    'run_test_config',
    'get_trainer_from_config',
    'get_paths_from_config',
    'get_callbacks_from_config',
    'get_trainer_by_name',
    'get_callbacks_by_name',
    'convert_pt_to_safetensors',
    'load_config',
]



import torch
import scipy
import numpy as np
from functools import lru_cache
from typing import Tuple

from reflectorch.data_generation.reflectivity.abeles import abeles

#Pytorch version based on the JAX implementation of pointwise smearing in the refnx package.

@lru_cache(maxsize=128)
def gauss_legendre(n: int) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Calculate Gaussian quadrature abscissae and weights.

    Args:
        n (int): Gaussian quadrature order.

    Returns:
        Tuple[torch.Tensor, torch.Tensor]: The abscissae and weights for Gauss-Legendre integration.
    """
    return scipy.special.p_roots(n)

def gauss(x: torch.Tensor) -> torch.Tensor:
    """
    Calculate the Gaussian function.

    Args:
        x (torch.Tensor): Input tensor.

    Returns:
        torch.Tensor: Output tensor after applying the Gaussian function.
    """
    return torch.exp(-0.5 * x * x)

def abeles_pointwise_smearing(
    q: torch.Tensor,
    dq: torch.Tensor,
    thickness: torch.Tensor,
    roughness: torch.Tensor,
    sld: torch.Tensor,
    gauss_num: int = 17,
    abeles_func=None,
    **abeles_kwargs,
) -> torch.Tensor:
    """
    Compute reflectivity with variable smearing using Gaussian quadrature.

    Args:
        q (torch.Tensor): The momentum transfer (q) values.
        dq (torch.Tensor): The resolution for curve smearing.
        thickness (torch.Tensor): The layer thicknesses.
        roughness (torch.Tensor): The interlayer roughnesses.
        sld (torch.Tensor): The SLDs of the layers.
        sld_magnetic (torch.Tensor, optional): The magnetic SLDs of the layers.
        magnetization_angle (torch.Tensor, optional): The magnetization angles.
        polarizer_eff (torch.Tensor, optional): The polarizer efficiency.
        analyzer_eff (torch.Tensor, optional): The analyzer efficiency.
        abeles_func (Callable, optional): A function implementing the simulation of the reflectivity curves.
        gauss_num (int, optional): Gaussian quadrature order. Defaults to 17.

    Returns:
        torch.Tensor: The computed reflectivity curves.
    """
    abeles_func = abeles_func or abeles

    if q.shape[0] == 1:
        q = q.repeat(thickness.shape[0], 1)

    _FWHM = 2 * np.sqrt(2 * np.log(2.0))
    _INTLIMIT = 3.5

    bs = q.shape[0]
    nq = q.shape[-1]
    device = q.device

    quad_order = gauss_num
    abscissa, weights = gauss_legendre(quad_order)
    abscissa = torch.tensor(abscissa)[None, :, None].to(device)
    weights = torch.tensor(weights)[None, :, None].to(device)
    prefactor = 1.0 / np.sqrt(2 * np.pi)

    gaussvals = prefactor * gauss(abscissa * _INTLIMIT)

    va = q[:, None, :] - _INTLIMIT * dq[:, None, :] / _FWHM
    vb = q[:, None, :] + _INTLIMIT * dq[:, None, :] / _FWHM

    qvals_for_res_0 = (abscissa * (vb - va) + vb + va) / 2
    qvals_for_res = qvals_for_res_0.reshape(bs, -1)

    refl_curves = abeles_func(
        q=qvals_for_res,
        thickness=thickness,
        roughness=roughness,
        sld=sld,
        **abeles_kwargs
    )

    # Handle multiple channels
    if refl_curves.dim() == 3:
        n_channels = refl_curves.shape[1]
        refl_curves = refl_curves.reshape(bs, n_channels, quad_order, nq)
        refl_curves = refl_curves * gaussvals.unsqueeze(1) * weights.unsqueeze(1)
        refl_curves = torch.sum(refl_curves, dim=2) * _INTLIMIT
    else:
        refl_curves = refl_curves.reshape(bs, quad_order, nq)
        refl_curves = refl_curves * gaussvals * weights
        refl_curves = torch.sum(refl_curves, dim=1) * _INTLIMIT

    return refl_curves


import numpy as np
import torch
import torch.nn.functional as F
from torch import nn
from dataclasses import dataclass
from typing import Optional

from reflectorch.data_generation import BATCH_DATA_TYPE
from reflectorch.ml.basic_trainer import Trainer
from reflectorch.ml.dataloaders import ReflectivityDataLoader

__all__ = [
    'RealTimeSimTrainer',
    'DenoisingAETrainer',
    'PointEstimatorTrainer',
]


@dataclass
class BasicBatchData:
    scaled_curves: torch.Tensor
    scaled_bounds: torch.Tensor
    scaled_params: torch.Tensor = None
    scaled_sigmas: Optional[torch.Tensor] = None
    scaled_q_values: Optional[torch.Tensor] = None
    scaled_denoised_curves: Optional[torch.Tensor] = None
    key_padding_mask: Optional[torch.Tensor] = None
    scaled_conditioning_params: Optional[torch.Tensor] = None
    unscaled_q_values: Optional[torch.Tensor] = None

class RealTimeSimTrainer(Trainer):
    """Trainer with functionality to customize the sampled batch of data"""
    loader: ReflectivityDataLoader

    def get_batch_by_idx(self, batch_num: int):
        """Gets a batch of data with the default batch size"""
        batch_data = self.loader.get_batch(self.batch_size)
        return self._get_batch(batch_data)

    def get_batch_by_size(self, batch_size: int):
        """Gets a batch of data with a custom batch size"""
        batch_data = self.loader.get_batch(batch_size)
        return self._get_batch(batch_data)

    def _get_batch(self, batch_data: BATCH_DATA_TYPE):
        """Modify the batch of data sampled from the data loader"""
        raise NotImplementedError


class PointEstimatorTrainer(RealTimeSimTrainer):
    """Point estimator trainer for the inverse problem."""

    def init(self):
        if getattr(self, 'use_l1_loss', False):
            self.criterion = nn.L1Loss(reduction='none')
        else:
            self.criterion = nn.MSELoss(reduction='none')
        self.use_curve_reconstruction_loss = getattr(self, 'use_curve_reconstruction_loss', False)
        self.rescale_loss_interval_width = getattr(self, 'rescale_loss_interval_width', False)
        if self.use_curve_reconstruction_loss:
            self.loader.calc_denoised_curves = True

        self.train_with_q_input = getattr(self, 'train_with_q_input', False)
        self.train_with_sigmas = getattr(self, 'train_with_sigmas', False)
        self.condition_on_q_resolutions = getattr(self, 'condition_on_q_resolutions', False)
    
    def _get_batch(self, batch_data: BATCH_DATA_TYPE) -> BasicBatchData:
        def get_scaled_or_none(key, scaler=None):
            value = batch_data.get(key)
            if value is None:
                return None
            scale_func = scaler or (lambda x: x)
            return scale_func(value).to(torch.float32)

        scaled_params = batch_data['scaled_params'].to(torch.float32)
        scaled_curves = batch_data['scaled_noisy_curves'].to(torch.float32)
        scaled_denoised_curves = get_scaled_or_none('curves', self.loader.curves_scaler.scale)
        scaled_q_values = get_scaled_or_none('q_values', self.loader.q_generator.scale_q) if self.train_with_q_input else None
        key_padding_mask = batch_data.get('key_padding_mask', None)

        scaled_q_resolutions = get_scaled_or_none('q_resolutions', self.loader.smearing.scale_resolutions) if self.condition_on_q_resolutions else None
        conditioning_params = []
        if scaled_q_resolutions is not None:
            conditioning_params.append(scaled_q_resolutions)
        scaled_conditioning_params = torch.cat(conditioning_params, dim=-1) if len(conditioning_params) > 0 else None
    
        num_params = scaled_params.shape[-1] // 3
        assert num_params * 3 == scaled_params.shape[-1]
        scaled_params, scaled_bounds = torch.split(scaled_params, [num_params, 2 * num_params], dim=-1)

        return BasicBatchData(
            scaled_params=scaled_params,
            scaled_bounds=scaled_bounds,
            scaled_curves=scaled_curves,
            scaled_q_values=scaled_q_values,
            scaled_denoised_curves=scaled_denoised_curves,
            scaled_conditioning_params=scaled_conditioning_params,
            unscaled_q_values=batch_data['q_values'],
            key_padding_mask=key_padding_mask,
        )

    def get_loss_dict(self, batch_data: BasicBatchData):
        """Returns the regression loss"""
        scaled_params=batch_data.scaled_params
        scaled_curves=batch_data.scaled_curves
        scaled_bounds=batch_data.scaled_bounds
        scaled_q_values=batch_data.scaled_q_values
        key_padding_mask=batch_data.key_padding_mask
        scaled_conditioning_params=batch_data.scaled_conditioning_params
        unscaled_q_values=batch_data.unscaled_q_values

        predicted_params = self.model(
            curves = scaled_curves,
            bounds = scaled_bounds,
            q_values = scaled_q_values,
            conditioning_params = scaled_conditioning_params,
            key_padding_mask = key_padding_mask,
            unscaled_q_values = unscaled_q_values,
        )

        if not self.rescale_loss_interval_width:
            loss = self.criterion(predicted_params, scaled_params).mean()
        else:
            n_params = scaled_params.shape[-1]
            b_min = scaled_bounds[..., :n_params]
            b_max = scaled_bounds[..., n_params:]
            interval_width = b_max - b_min

            base_loss  = self.criterion(predicted_params, scaled_params)
            if isinstance(self.criterion, torch.nn.MSELoss):
                width_factors = (interval_width / 2) ** 2
            elif isinstance(self.criterion, torch.nn.L1Loss):
                width_factors = interval_width / 2

            loss = (width_factors * base_loss).mean()

        return {'loss': loss}


# class PointEstimatorTrainer(RealTimeSimTrainer):
#     """Trainer for the regression inverse problem with incorporation of prior bounds"""
#     add_sigmas_to_context: bool = False
        
#     def _get_batch(self, batch_data: BATCH_DATA_TYPE):
#         scaled_params = batch_data['scaled_params'].to(torch.float32)
#         scaled_curves = batch_data['scaled_noisy_curves'].to(torch.float32)
#         if self.train_with_q_input:
#             q_values = batch_data['q_values'].to(torch.float32)
#             scaled_q_values = self.loader.q_generator.scale_q(q_values)
#         else:
#             scaled_q_values = None

#         num_params = scaled_params.shape[-1] // 3
#         assert num_params * 3 == scaled_params.shape[-1]
#         scaled_params, scaled_bounds = torch.split(scaled_params, [num_params, 2 * num_params], dim=-1)

#         return scaled_params, scaled_bounds, scaled_curves, scaled_q_values

#     def get_loss_dict(self, batch_data):
#         """computes the loss dictionary"""

#         scaled_params, scaled_bounds, scaled_curves, scaled_q_values = batch_data

#         if self.train_with_q_input:
#             predicted_params = self.model(scaled_curves, scaled_bounds, scaled_q_values)
#         else:
#             predicted_params = self.model(scaled_curves, scaled_bounds)
            
#         loss = self.mse(predicted_params, scaled_params)
#         return {'loss': loss}

#     def init(self):
#         self.mse = nn.MSELoss()


class DenoisingAETrainer(RealTimeSimTrainer):
    """Trainer which can be used for training a denoising autoencoder model. Overrides _get_batch and get_loss_dict methods """
    def init(self):
        self.loader.calc_denoised_curves = True
    
        if getattr(self, 'use_l1_loss', False):
            self.criterion = nn.L1Loss()
        else:
            self.criterion = nn.MSELoss()

    def _get_batch(self, batch_data: BATCH_DATA_TYPE):
        """returns scaled curves with and without noise"""
        scaled_noisy_curves, curves = batch_data['scaled_noisy_curves'], batch_data['curves']
        scaled_curves = self.loader.curves_scaler.scale(curves)

        scaled_noisy_curves, scaled_curves = scaled_noisy_curves.to(torch.float32), scaled_curves.to(torch.float32)

        return scaled_noisy_curves, scaled_curves

    def get_loss_dict(self, batch_data):
        """returns the reconstruction loss of the autoencoder"""
        scaled_noisy_curves, scaled_curves = batch_data
        restored_curves = self.model(scaled_noisy_curves)
        loss = self.criterion(scaled_curves, restored_curves)
        return {'loss': loss}
    


from typing import Tuple

import torch
from torch import Tensor

__all__ = [
    "MULTILAYER_MODELS",
    "MultilayerModel",
]


class MultilayerModel(object):
    NAME: str = ''
    PARAMETER_NAMES: Tuple[str, ...]

    def __init__(self, max_num_layers: int):
        self.max_num_layers = max_num_layers

    def to_standard_params(self, parametrized_model: Tensor) -> dict:
        raise NotImplementedError

    def from_standard_params(self, params: dict) -> Tensor:
        raise NotImplementedError


class BasicMultilayerModel1(MultilayerModel):
    NAME = 'repeating_multilayer_v1'

    PARAMETER_NAMES = (
        "d_full_rel",
        "rel_sigmas",
        "d_block",
        "s_block_rel",
        "r_block",
        "dr",
        "d3_rel",
        "s3_rel",
        "r3",
        "d_sio2",
        "s_sio2",
        "s_si",
        "r_sio2",
        "r_si",
    )

    def to_standard_params(self, parametrized_model: Tensor) -> dict:
        return multilayer_model1(parametrized_model, self.max_num_layers)


class BasicMultilayerModel2(MultilayerModel):
    NAME = 'repeating_multilayer_v2'

    PARAMETER_NAMES = (
        "d_full_rel",
        "rel_sigmas",
        "dr_sigmoid_rel_pos",
        "dr_sigmoid_rel_width",
        "d_block",
        "s_block_rel",
        "r_block",
        "dr",
        "d3_rel",
        "s3_rel",
        "r3",
        "d_sio2",
        "s_sio2",
        "s_si",
        "r_sio2",
        "r_si",
    )

    def to_standard_params(self, parametrized_model: Tensor) -> dict:
        return multilayer_model2(parametrized_model, self.max_num_layers)


class BasicMultilayerModel3(MultilayerModel):
    NAME = 'repeating_multilayer_v3'

    PARAMETER_NAMES = (
        "d_full_rel",
        "rel_sigmas",
        "dr_sigmoid_rel_pos",
        "dr_sigmoid_rel_width",
        "d_block1_rel",
        "d_block",
        "s_block_rel",
        "r_block",
        "dr",
        "d3_rel",
        "s3_rel",
        "r3",
        "d_sio2",
        "s_sio2",
        "s_si",
        "r_sio2",
        "r_si",
    )

    def to_standard_params(self, parametrized_model: Tensor) -> dict:
        return multilayer_model3(parametrized_model, self.max_num_layers)


MULTILAYER_MODELS = {
    'repeating_multilayer_v1': BasicMultilayerModel1,
    'repeating_multilayer_v2': BasicMultilayerModel2,
    'repeating_multilayer_v3': BasicMultilayerModel3,
}


def multilayer_model1(parametrized_model: Tensor, d_full_rel_max: int = 50) -> dict:
    n = d_full_rel_max

    (
        d_full_rel,
        rel_sigmas,
        d_block,
        s_block_rel,
        r_block,
        dr,
        d3_rel,
        s3_rel,
        r3,
        d_sio2,
        s_sio2,
        s_si,
        r_sio2,
        r_si,
    ) = parametrized_model.T

    batch_size = parametrized_model.shape[0]

    r_positions = 2 * n - torch.arange(2 * n, dtype=dr.dtype, device=dr.device)[None].repeat(batch_size, 1)

    r_modulations = torch.sigmoid(-(r_positions - 2 * d_full_rel[..., None]) / rel_sigmas[..., None])

    r_block = r_block[:, None].repeat(1, n)
    dr = dr[:, None].repeat(1, n)

    sld_blocks = torch.stack([r_block, r_block + dr], -1).flatten(1)

    sld_blocks = r_modulations * sld_blocks

    d3 = d3_rel * d_block

    thicknesses = torch.cat(
        [(d_block / 2)[:, None].repeat(1, n * 2), d3[:, None], d_sio2[:, None]], -1
    )

    s_block = s_block_rel * d_block

    roughnesses = torch.cat(
        [s_block[:, None].repeat(1, n * 2), (s3_rel * d3)[:, None], s_sio2[:, None], s_si[:, None]], -1
    )

    slds = torch.cat(
        [sld_blocks, r3[:, None], r_sio2[:, None], r_si[:, None]], -1
    )

    params = dict(
        thicknesses=thicknesses,
        roughnesses=roughnesses,
        slds=slds
    )
    return params


def multilayer_model2(parametrized_model: Tensor, d_full_rel_max: int = 50) -> dict:
    n = d_full_rel_max

    (
        d_full_rel,
        rel_sigmas,
        dr_sigmoid_rel_pos,
        dr_sigmoid_rel_width,
        d_block,
        s_block_rel,
        r_block,
        dr,
        d3_rel,
        s3_rel,
        r3,
        d_sio2,
        s_sio2,
        s_si,
        r_sio2,
        r_si,
    ) = parametrized_model.T

    batch_size = parametrized_model.shape[0]

    r_positions = 2 * n - torch.arange(2 * n, dtype=dr.dtype, device=dr.device)[None].repeat(batch_size, 1)

    r_modulations = torch.sigmoid(-(r_positions - 2 * d_full_rel[..., None]) / rel_sigmas[..., None])

    r_block = r_block[:, None].repeat(1, n)
    dr = dr[:, None].repeat(1, n)

    dr_positions = r_positions[:, ::2]

    dr_modulations = torch.sigmoid(
        -(dr_positions - (2 * d_full_rel * dr_sigmoid_rel_pos)[..., None]) / dr_sigmoid_rel_width[..., None]
    )

    dr = dr * dr_modulations

    sld_blocks = torch.stack([r_block, r_block + dr], -1).flatten(1)

    sld_blocks = r_modulations * sld_blocks

    d3 = d3_rel * d_block

    thicknesses = torch.cat(
        [(d_block / 2)[:, None].repeat(1, n * 2), d3[:, None], d_sio2[:, None]], -1
    )

    s_block = s_block_rel * d_block

    roughnesses = torch.cat(
        [s_block[:, None].repeat(1, n * 2), (s3_rel * d3)[:, None], s_sio2[:, None], s_si[:, None]], -1
    )

    slds = torch.cat(
        [sld_blocks, r3[:, None], r_sio2[:, None], r_si[:, None]], -1
    )

    params = dict(
        thicknesses=thicknesses,
        roughnesses=roughnesses,
        slds=slds
    )
    return params


def multilayer_model3(parametrized_model: Tensor, d_full_rel_max: int = 30):
    n = d_full_rel_max

    (
        d_full_rel,
        rel_sigmas,
        dr_sigmoid_rel_pos,
        dr_sigmoid_rel_width,
        d_block1_rel,
        d_block,
        s_block_rel,
        r_block,
        dr,
        d3_rel,
        s3_rel,
        r3,
        d_sio2,
        s_sio2,
        s_si,
        r_sio2,
        r_si,
    ) = parametrized_model.T

    batch_size = parametrized_model.shape[0]

    r_positions = 2 * n - torch.arange(2 * n, dtype=dr.dtype, device=dr.device)[None].repeat(batch_size, 1)

    r_modulations = torch.sigmoid(
        -(
                r_positions - 2 * d_full_rel[..., None]
        ) / rel_sigmas[..., None]
    )

    dr_positions = r_positions[:, ::2]

    dr_modulations = dr[..., None] * (1 - torch.sigmoid(
        -(
                dr_positions - 2 * d_full_rel[..., None] + 2 * dr_sigmoid_rel_pos[..., None]
        ) / dr_sigmoid_rel_width[..., None]
    ))

    r_block = r_block[..., None].repeat(1, n)
    dr = dr[..., None].repeat(1, n)

    sld_blocks = torch.stack(
        [
            r_block + dr_modulations * (1 - d_block1_rel[..., None]),
            r_block + dr - dr_modulations * d_block1_rel[..., None]
        ], -1).flatten(1)

    sld_blocks = r_modulations * sld_blocks

    d3 = d3_rel * d_block

    d1, d2 = d_block * d_block1_rel, d_block * (1 - d_block1_rel)

    thickness_blocks = torch.stack([d1[:, None].repeat(1, n), d2[:, None].repeat(1, n)], -1).flatten(1)

    thicknesses = torch.cat(
        [thickness_blocks, d3[:, None], d_sio2[:, None]], -1
    )

    s_block = s_block_rel * d_block

    roughnesses = torch.cat(
        [s_block[:, None].repeat(1, n * 2), (s3_rel * d3)[:, None], s_sio2[:, None], s_si[:, None]], -1
    )

    slds = torch.cat(
        [sld_blocks, r3[:, None], r_sio2[:, None], r_si[:, None]], -1
    )

    params = dict(
        thicknesses=thicknesses,
        roughnesses=roughnesses,
        slds=slds
    )
    return params



import math
from torch.optim import lr_scheduler

import numpy as np

from reflectorch.ml.basic_trainer import Trainer, TrainerCallback, PeriodicTrainerCallback

__all__ = [
    'ScheduleBatchSize',
    'ScheduleLR',
    'StepLR',
    'CyclicLR',
    'LogCyclicLR',
    'ReduceLROnPlateau',
    'OneCycleLR',
    'CosineAnnealingWithWarmup',
]


class ScheduleBatchSize(PeriodicTrainerCallback):
    """Batch size scheduler

    Args:
        step (int): number of iterations after which the batch size is modified.
        gamma (int, optional): quantity which is added to or multiplied with the current batch size. Defaults to 2.
        last_epoch (int, optional): the last training iteration for which the batch size is modified. Defaults to -1.
        mode (str, optional): ``'add'`` for addition or ``'multiply'`` for multiplication. Defaults to 'add'.
    """
    def __init__(self, step: int, gamma: int = 2, last_epoch: int = -1, mode: str = 'add'):
        super().__init__(step, last_epoch)

        assert mode in ('add', 'multiply')

        self.gamma = gamma
        self.mode = mode

    def _end_batch(self, trainer: Trainer, batch_num: int) -> None:
        if self.mode == 'add':
            trainer.batch_size += self.gamma
        elif self.mode == 'multiply':
            trainer.batch_size *= self.gamma


class ScheduleLR(TrainerCallback):
    """Base class for learning rate schedulers

    Args:
        lr_scheduler_cls: class of the learning rate scheduler
    """
    
    def __init__(self, lr_scheduler_cls, **kwargs):
        self.lr_scheduler_cls = lr_scheduler_cls
        self.kwargs = kwargs
        self.lr_scheduler = None

    def start_training(self, trainer: Trainer) -> None:
        """initializes a learning rate scheduler based on its class and keyword arguments at the start of training

        Args:
            trainer (Trainer): the trainer object
        """
        self.lr_scheduler = self.lr_scheduler_cls(trainer.optim, **self.kwargs)
        trainer.callback_params['lrs'] = []

    def end_batch(self, trainer: Trainer, batch_num: int) -> None:
        """modifies the learning rate at the end of each iteration

        Args:
            trainer (Trainer): the trainer object
            batch_num (int): index of the current iteration
        """
        self.lr_scheduler.step()

    def simulate_and_plot(self, total_steps: int, initial_lr: float, log_scale: bool = False):
        import torch
        import matplotlib.pyplot as plt

        dummy_optim = torch.optim.Adam([torch.zeros(1)], lr=initial_lr)
        scheduler = self.lr_scheduler_cls(dummy_optim, **self.kwargs)

        lrs = []
        for step in range(total_steps):
            lrs.append(dummy_optim.param_groups[0]['lr'])
            scheduler.step()

        plt.figure(figsize=(10, 6))
        plt.plot(lrs, label='Learning Rate')
        plt.xlabel('Steps')
        plt.ylabel('Learning Rate')
        plt.title('Learning Rate Schedule')

        if log_scale:
            plt.yscale('log')

        plt.grid(True, which="both", linestyle='--', linewidth=0.5)
        plt.legend()
        plt.show()


class StepLR(ScheduleLR):
    """Learning rate scheduler which decays the learning rate of each parameter group by gamma every ``step_size`` epochs.
    
    Args:
        step_size (int): Period of learning rate decay
        gamma (float): Multiplicative factor of learning rate decay
        last_epoch (int, optional): The index of last iteration. Defaults to -1.
    """
    def __init__(self, step_size: int, gamma: float, last_epoch: int = -1, **kwargs):

        
        super().__init__(lr_scheduler.StepLR, step_size=step_size, gamma=gamma, last_epoch=last_epoch, **kwargs)

    def start_training(self, trainer: Trainer) -> None:
        trainer.optim.param_groups[0]['initial_lr'] = trainer.lr()
        super().start_training(trainer)


class CyclicLR(ScheduleLR):
    """Cyclic learning rate scheduler

    Args:
        base_lr (float): Initial learning rate which is the lower boundary in the cycle
        max_lr (float): Upper learning rate boundary in the cycle
        step_size_up (int, optional): Number of training iterations in the increasing half of a cycle. Defaults to 2000.
        cycle_momentum (bool, optional):  If True, momentum is cycled inversely to learning rate between ``base_momentum`` and ``max_momentum``. Defaults to False.
        gamma (float, optional): Constant in ``‘exp_range’`` mode scaling function: gamma^(cycle iterations). Defaults to 1.
        mode (str, optional):  One of: ``'triangular'`` (a basic triangular cycle without amplitude scaling),
              ``'triangular2'`` (a basic triangular cycle that scales initial amplitude by half each cycle), ``'exp_range'`` 
              (a cycle that scales initial amplitude by gamma^iterations at each cycle iteration). Defaults to 'triangular'.
    """
    def __init__(self, base_lr, max_lr, step_size_up: int = 2000,
                 cycle_momentum: bool = False, gamma: float = 1., mode: str = 'triangular',
                 **kwargs):
        super().__init__(
            lr_scheduler.CyclicLR,
            base_lr=base_lr,
            max_lr=max_lr,
            step_size_up=step_size_up,
            cycle_momentum=cycle_momentum,
            gamma=gamma,
            mode=mode,
            **kwargs
        )


class LogCyclicLR(TrainerCallback):
    """Cyclic learning rate scheduler on a logarithmic scale
    
    Args:
        base_lr (float): Lower learning rate boundary in the cycle
        max_lr (float): Upper learning rate boundary in the cycle
        period (int, optional): Number of training iterations in the cycle. Defaults to 2000.
        gamma (float, optional): Constant for scaling the amplitude as ``gamma`` ^ ``iterations``. Defaults to 1.
        start_period (int, optional): Number of starting iterations with the default learning rate.
        log (bool, optional): If ``True``, the cycle is in the logarithmic domain.
        param_groups (tupe, optional): Parameter groups of the optimizer.
    """
    def __init__(self,
                 base_lr,
                 max_lr,
                 period: int = 2000,
                 gamma: float = None,
                 log: bool = True,
                 param_groups: tuple = (0,),
                 start_period: int = 25,
                 ):
        self.base_lr = base_lr
        self.max_lr = max_lr
        self.period = period
        self.gamma = gamma
        self.param_groups = param_groups
        self.log = log
        self.start_period = start_period
        self._axis = None
        self._period = None

    def get_lr(self, batch_num: int):
        return self._get_lr(batch_num)

    def _get_lr(self, batch_num):
        num_period, t = batch_num // self.period, batch_num % self.period

        if self._period != num_period:
            self._period = num_period
            if self.gamma and (num_period >= self.start_period):
                amp = (self.max_lr - self.base_lr) * (self.gamma ** (num_period - self.start_period))
                max_lr = self.base_lr + amp
            else:
                max_lr = self.max_lr

            if self.log:
                self._axis = np.logspace(np.log10(self.base_lr), np.log10(max_lr), self.period // 2)
            else:
                self._axis = np.linspace(self.base_lr, max_lr, self.period // 2)
        if t < self.period // 2:
            lr = self._axis[t]
        else:
            lr = self._axis[self.period - t - 1]
        return lr

    def end_batch(self, trainer: Trainer, batch_num: int):
        lr = self.get_lr(batch_num)
        for param_group in self.param_groups:
            trainer.set_lr(lr, param_group)

    def simulate_and_plot(self, total_steps: int, log_scale: bool = True):
        import matplotlib.pyplot as plt
        lrs = [self.get_lr(batch_num) for batch_num in range(total_steps)]

        plt.figure(figsize=(10, 6))
        plt.plot(lrs, label='Learning Rate')
        plt.xlabel('Steps')
        plt.ylabel('Learning Rate')
        plt.title('Learning Rate Schedule')
        if log_scale:
            plt.yscale('log')
        plt.grid(True, which='both', linestyle='--', linewidth=0.5)
        plt.legend()
        plt.show()


class ReduceLROnPlateau(TrainerCallback):
    """Learning rate scheduler which reduces the learning rate when the loss stops decreasing
    
    Args:
            gamma (float, optional): Multiplicative factor of learning rate decay. Defaults to 0.5.
            patience (int, optional): The number of allowed iterations with no improvement after which the learning rate will be reduced. Defaults to 500.
            average (int, optional): Size of the window over which the average loss is computed. Defaults to 50.
            loss_key (str, optional): Defaults to 'total_loss'.
            param_groups (tuple, optional): Defaults to (0,).
    
    """
    def __init__(
            self,
            gamma: float = 0.5,
            patience: int = 500,
            average: int = 50,
            loss_key: str = 'total_loss',
            param_groups: tuple = (0,),
    ):
        """
        """
        self.patience = patience
        self.average = average
        self.gamma = gamma
        self.loss_key = loss_key
        self.param_groups = param_groups

    def end_batch(self, trainer: Trainer, batch_num: int) -> None:
        loss = trainer.losses[self.loss_key]

        if len(loss) < self.patience:
            return

        if np.mean(loss[-self.patience:-(self.patience - self.average)]) <= np.mean(loss[-self.average:]):
            for param_group in self.param_groups:
                trainer.set_lr(trainer.lr(param_group) * self.gamma, param_group)
                
                
class OneCycleLR(ScheduleLR):
    """One-cycle learning rate scheduler (https://arxiv.org/abs/1708.07120)

    Args:
        max_lr (float): Upper learning rate boundary in the cycle
        total_steps (int): The total number of steps in the cycle
        pct_start (float, optional): The percentage of the cycle (in number of steps) spent increasing the learning rate. Defaults to 0.3.
        div_factor (float, optional): Determines the initial learning rate via initial_lr = ``max_lr`` / ``div_factor``. Defaults to 25..
        final_div_factor (float, optional): Determines the minimum learning rate via min_lr = ``initial_lr`` / ``final_div_factor``. Defaults to 1e4.
        three_phase (bool, optional): If ``True``, use a third phase of the schedule to annihilate the learning rate according to ``final_div_factor`` instead of modifying the second phase. Defaults to True.
    """
    def __init__(self, max_lr: float, total_steps: int, pct_start: float = 0.3, div_factor: float = 25., 
                 final_div_factor: float = 1e4, three_phase: bool = True, **kwargs):
        super().__init__(
            lr_scheduler.OneCycleLR,
            max_lr=max_lr,
            total_steps=total_steps,
            pct_start=pct_start,
            div_factor=div_factor ,
            final_div_factor=final_div_factor,
            three_phase=three_phase,
            **kwargs
        )

class CosineAnnealingWithWarmup(TrainerCallback):
    """
    Cosine annealing scheduler with a warm-up stage.

    Args:
        max_lr (float): The maximum learning rate after the warm-up phase.
        min_lr (float): The minimum learning rate after the warm-up phase.
        warmup_iters (int): The number of iterations for the warm-up phase.
        total_iters (int): The total number of iterations for the scheduler (including warm-up).
    """
    def __init__(self, max_lr=None, min_lr=1.0e-6, warmup_iters=100, total_iters=100000):
        self.max_lr = max_lr
        self.min_lr = min_lr
        self.warmup_iters = warmup_iters
        self.total_iters = total_iters

    def get_lr(self, step):
        """
        Compute the learning rate for a given iteration.

        Args:
            step (int): The current iteration.

        Returns:
            float: The learning rate for the current iteration.
        """
        if step < self.warmup_iters:
            # Warm-up stage: Linear increase from 0 to max_lr
            return self.max_lr * step / self.warmup_iters
        elif step < self.total_iters:
            # Cosine annealing stage
            t = (step - self.warmup_iters) / (self.total_iters - self.warmup_iters)
            return self.min_lr + 0.5 * (self.max_lr - self.min_lr) * (1 + math.cos(math.pi * t))
        else:
            # Beyond total iterations: Return min_lr
            return self.min_lr
    
    def start_training(self, trainer: Trainer) -> None:
        self.max_lr = trainer.lr()

    def end_batch(self, trainer: Trainer, batch_num: int):
        """
        Updates the learning rate at the end of each batch.

        Args:
            trainer (Trainer): The trainer object.
            batch_num (int): The current batch number.
        """
        lr = self.get_lr(batch_num)
        trainer.set_lr(lr)

    def simulate_and_plot(self, total_steps: int = None, log_scale: bool = False):
        """
        Simulates and plots the learning rate evolution.

        Args:
            total_batches (int, optional): Total number of batches to simulate. If None, uses self.total_iters.
        """
        
        total_steps = total_steps or self.total_iters
        lrs = [self.get_lr(step) for step in range(total_steps)]

        import matplotlib.pyplot as plt
        plt.figure(figsize=(10, 6))
        plt.plot(lrs, label='Learning Rate')
        plt.xlabel('Steps')
        plt.ylabel('Learning Rate')
        plt.title('Learning Rate Scheduler')
        if log_scale:
            plt.yscale('log')
        plt.grid(True, which='both', linestyle='--', linewidth=0.5)
        plt.legend()
        plt.show()


import numpy as np


def interp_reflectivity(q_interp, q, reflectivity, min_value: float = 1e-10, logspace = False):
    """Interpolate data on a base 10 logarithmic scale

    Args:
        q_interp (array-like): reciprocal space points used for the interpolation
        q (array-like): reciprocal space points of the measured reflectivity curve
        reflectivity (array-like): reflectivity curve measured at the points ``q``
        min_value (float, optional): minimum intensity of the reflectivity curve. Defaults to 1e-10.

    Returns:
        array-like: interpolated reflectivity curve
    """
    if not(logspace):
        return 10 ** np.interp(q_interp, q, np.log10(np.clip(reflectivity, min_value, None)))
    else:
        return 10 ** np.interp(np.log10(q_interp), np.log10(q), np.log10(np.clip(reflectivity, min_value, None)))



import matplotlib.pyplot as plt


def plot_losses(
        losses: dict,
        log: bool = False,
        show: bool = True,
        title: str = 'Losses',
        x_label: str = 'Iterations',
        best_epoch: float = None,
        **kwargs
):
    func = plt.semilogy if log else plt.plot

    if len(losses) <= 2:
        losses = {'loss': losses['total_loss']}

    for k, data in losses.items():
        func(data, label=k, **kwargs)

    if best_epoch is not None:
        plt.axvline(best_epoch, ls='--', color='red')

    plt.xlabel(x_label)

    if len(losses) > 2:
        plt.legend()

    plt.title(title)

    if show:
        plt.show()



import logging
from pathlib import Path

import torch
from torch import nn, load

from reflectorch.models.activations import activation_by_name
from reflectorch.paths import SAVED_MODELS_DIR

__all__ = [
    "ConvEncoder",
    "ConvDecoder",
    "ConvAutoencoder",
]

logger = logging.getLogger(__name__)


class ConvEncoder(nn.Module):
    """A 1D CNN encoder / embedding network

    Args:
        in_channels (int, optional): the number of input channels. Defaults to 1.
        hidden_channels (tuple, optional): the number of intermediate channels of each convolutional layer. Defaults to (32, 64, 128, 256, 512).
        dim_embedding (int, optional): the dimension of the output latent embedding. Defaults to 64.
        dim_avpool (int, optional): the output size of the adaptive average pooling layer. Defaults to 1.
        use_batch_norm (bool, optional): whether to use batch normalization. Defaults to True.
        activation (str, optional): the type of activation function. Defaults to 'relu'.
    """
    def __init__(self,
                 in_channels: int = 1,
                 hidden_channels: tuple = (32, 64, 128, 256, 512),
                 kernel_size: int = 3,
                 dim_embedding: int = 64,
                 dim_avpool: int = 1,
                 use_batch_norm: bool = True,
                 use_se: bool = False,
                 activation: str = 'relu',
                 ):
        super().__init__()

        modules = []

        activation = activation_by_name(activation)

        for h in hidden_channels:
            layers = [
                nn.Conv1d(in_channels, out_channels=h, kernel_size=kernel_size, stride=2, padding=kernel_size // 2),
                activation(),
            ]

            if use_batch_norm:
                layers.insert(1, nn.BatchNorm1d(h))

            if use_se:
                layers.insert(2, SEBlock(h))

            modules.append(nn.Sequential(*layers))
            in_channels = h

        self.core = nn.Sequential(*modules)
        self.avpool = nn.AdaptiveAvgPool1d(dim_avpool)
        self.fc = nn.Linear(hidden_channels[-1] * dim_avpool, dim_embedding)

    def forward(self, x):
        if len(x.shape) < 3:
            x = x.unsqueeze(1)
        x = self.core(x)
        x = self.avpool(x).view(x.size(0), -1)
        x = self.fc(x)
        return x

    def load_weights(self, path: str or Path = None, strict: bool = False):
        if not path:
            return

        if isinstance(path, str):
            if not path.endswith('.pt'):
                path = path + '.pt'
            path = SAVED_MODELS_DIR / path

        if not path.is_file():
            logger.error(f'File {str(path)} is not found.')
            return
        try:
            state_dict = load(path)
            self.load_state_dict(state_dict, strict=strict)
        except Exception as err:
            logger.exception(err)


class ConvDecoder(nn.Module):
    """A 1D CNN decoder

    Args:
        hidden_dims (tuple, optional): the number of intermediate channels of each convolutional layer. Defaults to (512, 256, 128, 64, 32).
        latent_dim (int, optional): the dimension of the input latent embedding. Defaults to 64.
        in_size (int, optional): the initial size for upscaling. Defaults to 8.
        use_batch_norm (bool, optional): whether to use batch normalization. Defaults to True.
        activation (str, optional): the type of activation function. Defaults to 'relu'.
    """
    def __init__(self, 
                 hidden_channels: tuple = (512, 256, 128, 64, 32), 
                 dim_latent: int = 64, 
                 in_size: int = 8,
                 kernel_size: int = 3,
                 use_batch_norm: bool = True,
                 activation: str = 'relu',
                 ):

        super().__init__()

        self.in_size = in_size
        modules = []

        self.decoder_input = nn.Linear(dim_latent, hidden_channels[0] * in_size)

        activation = activation_by_name(activation)

        for i in range(len(hidden_channels) - 1):
            modules.append(
                nn.Sequential(
                    nn.ConvTranspose1d(
                        hidden_channels[i],
                        hidden_channels[i + 1],
                        kernel_size=kernel_size, #3
                        stride=2,
                        padding=kernel_size // 2, #1
                        output_padding=1,
                    ),
                    nn.BatchNorm1d(hidden_channels[i + 1]) if use_batch_norm else nn.Identity(),
                    activation(),
                )
            )

        self.decoder = nn.Sequential(*modules)

        self.final_layer = nn.Sequential(
            nn.ConvTranspose1d(hidden_channels[-1],
                               hidden_channels[-1],
                               kernel_size=kernel_size, #3
                               stride=2,
                               padding=kernel_size // 2, #1
                               output_padding=1),
            nn.BatchNorm1d(hidden_channels[-1]) if use_batch_norm else nn.Identity(),
            activation(),
            nn.Conv1d(hidden_channels[-1], out_channels=1,
                      kernel_size=3, padding=1)
        )

    def forward(self, x):
        batch_size = x.shape[0]
        x = self.decoder_input(x).view(batch_size, -1, self.in_size)
        x = self.decoder(x)
        x = self.final_layer(x).flatten(1)
        return x


class ConvAutoencoder(nn.Module):
    """A 1D convolutional denoising autoencoder"""
    def __init__(self,
                 in_channels: int = 1,
                 encoder_hidden_channels: tuple = (32, 64, 128, 256, 512),
                 decoder_hidden_channels: tuple = (512, 256, 128, 64, 32),
                 dim_latent: int = 64,
                 dim_avpool: int = 1,
                 kernel_size: int = 3,
                 use_batch_norm: bool = True,
                 activation: str = 'relu',
                 decoder_in_size: int = 8,
                 **kwargs
                 ):
        super().__init__()
        self.encoder = ConvEncoder(
            in_channels=in_channels, 
            hidden_channels=encoder_hidden_channels,
            kernel_size=kernel_size,
            dim_embedding=dim_latent, 
            dim_avpool=dim_avpool, 
            use_batch_norm=use_batch_norm, 
            activation=activation,
            **kwargs)

        self.decoder = ConvDecoder(
            hidden_channels=decoder_hidden_channels, 
            dim_latent=dim_latent, 
            in_size=decoder_in_size, 
            kernel_size=kernel_size,
            use_batch_norm=use_batch_norm, 
            activation=activation, 
            **kwargs)

    def forward(self, x):
        return self.decoder(self.encoder(x))
    
class SEBlock(nn.Module):
    """Squeeze-and-excitation block (https://arxiv.org/abs/1709.01507) """
    def __init__(self, in_channels, reduction=16):
        super().__init__()
        self.fc1 = nn.Linear(in_channels, in_channels // reduction, bias=False)
        self.fc2 = nn.Linear(in_channels // reduction, in_channels, bias=False)
        self.relu = nn.ReLU()
        self.sigmoid = nn.Sigmoid()
        self.global_avg_pool = nn.AdaptiveAvgPool1d(1)

    def forward(self, x):
        batch_size, channels, _ = x.size()
        
        #Squeeze
        se = self.global_avg_pool(x).view(batch_size, channels)
        
        #Excitation
        se = self.fc1(se)
        se = self.relu(se)
        se = self.fc2(se)
        se = self.sigmoid(se).view(batch_size, channels, 1)
        
        #Scale the input feature maps (channel-wise attention)
        return x * se


from typing import Tuple, Union
from pathlib import Path
import subprocess

from reflectorch.paths import RUN_SCRIPTS_DIR


def save_sbatch_and_run(
        name: str,
        args: str,
        time: str,
        partition: str = None,
        reservation: bool = False,
        chdir: str = '~/maxwell_output',
        run_dir: Path = None,
        confirm: bool = False,
) -> Union[Tuple[str, str], None]:
    run_dir = Path(run_dir) if run_dir else RUN_SCRIPTS_DIR
    sbatch_path = run_dir / f'{name}.sh'

    if sbatch_path.is_file():
        import warnings
        warnings.warn(f'Sbatch file {str(sbatch_path)} already exists!')
        if confirm and not confirm_input('Continue?'):
            return

    file_content = _generate_sbatch_str(
        name,
        args,
        time=time,
        reservation=reservation,
        partition=partition,
        chdir=chdir,
    )

    if confirm:
        print(f'Generated file content: \n{file_content}\n')
        if not confirm_input(f'Save to {str(sbatch_path)} and run?'):
            return

    with open(str(sbatch_path), 'w') as f:
        f.write(file_content)

    res = submit_job(str(sbatch_path))
    return res


def _generate_sbatch_str(name: str,
                         args: str,
                         time: str,
                         partition: str = None,
                         reservation: bool = False,
                         chdir: str = '~/maxwell_output',
                         entry_point: str = 'python -m reflectorch.train',
                         ) -> str:
    chdir = str(Path(chdir).expanduser().absolute())
    partition_keyword = 'reservation' if reservation else 'partition'

    return f'''#!/bin/bash
#SBATCH --chdir {chdir}
#SBATCH --{partition_keyword}={partition}
#SBATCH --constraint=P100
#SBATCH --nodes=1
#SBATCH --job-name {name}
#SBATCH --time={time}
#SBATCH --output {name}.out
#SBATCH --error {name}.err

{entry_point} {args}
'''


def confirm_input(message: str) -> bool:
    yes = ('y', 'yes')
    no = ('n', 'no')
    res = ''
    valid_results = list(yes) + list(no)
    message = f'{message} Y/n: '

    while res not in valid_results:
        res = input(message).lower()
    return res in yes


def submit_job(sbatch_path: str) -> Tuple[str, str]:
    process = subprocess.Popen(
        ['sbatch', str(sbatch_path)],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    stdout, stderr = process.communicate()
    return stdout.decode(), stderr.decode()



from reflectorch.inference.inference_model import InferenceModel, EasyInferenceModel
from reflectorch.inference.query_matcher import HuggingfaceQueryMatcher
from reflectorch.inference.multilayer_inference_model import MultilayerInferenceModel
from reflectorch.inference.preprocess_exp import (
    StandardPreprocessing,
    standard_preprocessing,
    interp_reflectivity,
    apply_attenuation_correction,
    apply_footprint_correction,
)
from reflectorch.inference.torch_fitter import ReflGradientFit

__all__ = [
    "InferenceModel",
    "EasyInferenceModel",
    "MultilayerInferenceModel",
    "HuggingfaceQueryMatcher",
    "StandardPreprocessing",
    "standard_preprocessing",
    "ReflGradientFit",
    "interp_reflectivity",
    "apply_attenuation_correction",
    "apply_footprint_correction",
]



import torch
from torch import Tensor

from reflectorch.data_generation.utils import (
    uniform_sampler,
    logdist_sampler,
)

from reflectorch.data_generation.priors.utils import get_max_allowed_roughness


class SamplerStrategy(object):
    """Base class for sampler strategies"""
    def sample(self, batch_size: int,
               total_min_bounds: Tensor,
               total_max_bounds: Tensor,
               total_min_delta: Tensor,
               total_max_delta: Tensor,
               ):
        raise NotImplementedError


class BasicSamplerStrategy(SamplerStrategy):
    """Sampler strategy with no constraints on the values of the parameters

    Args:
        logdist (bool, optional): if True the relative widths of the subprior intervals are sampled uniformly on a logarithmic scale instead of uniformly. Defaults to False.
    """
    def __init__(self, logdist: bool = False):
        if logdist:
            self.widths_sampler_func = logdist_sampler
        else:
            self.widths_sampler_func = uniform_sampler

    def sample(self, batch_size: int,
               total_min_bounds: Tensor,
               total_max_bounds: Tensor,
               total_min_delta: Tensor,
               total_max_delta: Tensor,
               ):
        """
        Args:
            batch_size (int): the batch size
            total_min_bounds (Tensor): mimimum values of the parameters
            total_max_bounds (Tensor): maximum values of the parameters
            total_min_delta (Tensor): minimum widths of the subprior intervals
            total_max_delta (Tensor): maximum widths of the subprior intervals

        Returns:
            tuple(Tensor): samples the values of the parameters and their prior bounds (params, min_bounds, max_bounds). The widths W of the subprior interval are sampled first, then the centers C of the subprior interval, such that the prior bounds are C-W/2 and C+W/2, then the parameters are sampled from [C-W/2, C+W/2] )
        """
        return basic_sampler(
            batch_size,
            total_min_bounds,
            total_max_bounds,
            total_min_delta,
            total_max_delta,
            self.widths_sampler_func,
        )


class ConstrainedRoughnessSamplerStrategy(BasicSamplerStrategy):
    """Sampler strategy where the roughnesses are constrained not to exceed a fraction of the two neighboring thicknesses 

    Args:
        thickness_mask (Tensor): indices in the tensors which correspond to thicknesses
        roughness_mask (Tensor): indices in the tensors which correspond to roughnesses
        logdist (bool, optional): if ``True`` the relative widths of the subprior intervals are sampled uniformly on a logarithmic scale instead of uniformly. Defaults to False.
        max_thickness_share (float, optional): fraction of the layer thickness that the roughness should not exceed. Defaults to 0.5.
    """
    def __init__(self,
                 thickness_mask: Tensor,
                 roughness_mask: Tensor,
                 logdist: bool = False,
                 max_thickness_share: float = 0.5,
                 max_total_thickness: float = None,
                 ):
        super().__init__(logdist=logdist)
        self.thickness_mask = thickness_mask
        self.roughness_mask = roughness_mask
        self.max_thickness_share = max_thickness_share
        self.max_total_thickness = max_total_thickness

    def sample(self, batch_size: int,
               total_min_bounds: Tensor,
               total_max_bounds: Tensor,
               total_min_delta: Tensor,
               total_max_delta: Tensor,
               ):
        """
        Args:
            batch_size (int): the batch size
            total_min_bounds (Tensor): mimimum values of the parameters
            total_max_bounds (Tensor): maximum values of the parameters
            total_min_delta (Tensor): minimum widths of the subprior intervals
            total_max_delta (Tensor): maximum widths of the subprior intervals

        Returns:
            tuple(Tensor): samples the values of the parameters and their prior bounds *(params, min_bounds, max_bounds)*, the roughnesses being constrained. The widths **W** of the subprior interval are sampled first, then the centers **C** of the subprior interval, such that the prior bounds are **C** - **W** / 2 and **C** + **W** / 2, then the parameters are sampled from [**C** - **W** / 2, **C** + **W** / 2] )
        """
        device = total_min_bounds.device
        return constrained_roughness_sampler(
            batch_size,
            total_min_bounds,
            total_max_bounds,
            total_min_delta,
            total_max_delta,
            thickness_mask=self.thickness_mask.to(device),
            roughness_mask=self.roughness_mask.to(device),
            widths_sampler_func=self.widths_sampler_func,
            coef_roughness=self.max_thickness_share,
            max_total_thickness=self.max_total_thickness,
        )

class ConstrainedRoughnessAndImgSldSamplerStrategy(BasicSamplerStrategy):
    """Sampler strategy where the roughnesses are constrained not to exceed a fraction of the two neighboring thicknesses, and the imaginary slds are constrained not to exceed a fraction of the real slds

    Args:
        thickness_mask (Tensor): indices in the tensors which correspond to thicknesses
        roughness_mask (Tensor): indices in the tensors which correspond to roughnesses
        sld_mask (Tensor): indices in the tensors which correspond to real slds
        isld_mask (Tensor): indices in the tensors which correspond to imaginary slds
        logdist (bool, optional): if ``True`` the relative widths of the subprior intervals are sampled uniformly on a logarithmic scale instead of uniformly. Defaults to False.
        max_thickness_share (float, optional): fraction of the layer thickness that the roughness should not exceed. Defaults to 0.5
        max_sld_share (float, optional): fraction of the real sld that the imaginary sld should not exceed. Defaults to 0.2.
    """
    def __init__(self,
                 thickness_mask: Tensor,
                 roughness_mask: Tensor,
                 sld_mask: Tensor,
                 isld_mask: Tensor,
                 logdist: bool = False,
                 max_thickness_share: float = 0.5,
                 max_sld_share: float = 0.2,
                 max_total_thickness: float = None,
                 ):
        super().__init__(logdist=logdist)
        self.thickness_mask = thickness_mask
        self.roughness_mask = roughness_mask
        self.sld_mask = sld_mask
        self.isld_mask = isld_mask
        self.max_thickness_share = max_thickness_share
        self.max_sld_share = max_sld_share
        self.max_total_thickness = max_total_thickness

    def sample(self, batch_size: int,
               total_min_bounds: Tensor,
               total_max_bounds: Tensor,
               total_min_delta: Tensor,
               total_max_delta: Tensor,
               ):
        """
        Args:
            batch_size (int): the batch size
            total_min_bounds (Tensor): mimimum values of the parameters
            total_max_bounds (Tensor): maximum values of the parameters
            total_min_delta (Tensor): minimum widths of the subprior intervals
            total_max_delta (Tensor): maximum widths of the subprior intervals

        Returns:
            tuple(Tensor): samples the values of the parameters and their prior bounds *(params, min_bounds, max_bounds)*, the roughnesses and imaginary slds being constrained. The widths **W** of the subprior interval are sampled first, then the centers **C** of the subprior interval, such that the prior bounds are **C** - **W** /2 and **C** + **W** / 2, then the parameters are sampled from [**C** - **W** / 2, **C** + **W** / 2] )
        """
        device = total_min_bounds.device
        return constrained_roughness_and_isld_sampler(
            batch_size,
            total_min_bounds,
            total_max_bounds,
            total_min_delta,
            total_max_delta,
            thickness_mask=self.thickness_mask.to(device),
            roughness_mask=self.roughness_mask.to(device),
            sld_mask=self.sld_mask.to(device),
            isld_mask=self.isld_mask.to(device),
            widths_sampler_func=self.widths_sampler_func,
            coef_roughness=self.max_thickness_share,
            coef_isld=self.max_sld_share,
            max_total_thickness=self.max_total_thickness,
        )

def basic_sampler(
        batch_size: int,
        total_min_bounds: Tensor,
        total_max_bounds: Tensor,
        total_min_delta: Tensor,
        total_max_delta: Tensor,
        widths_sampler_func,
):

    delta_vector = total_max_bounds - total_min_bounds

    prior_widths = widths_sampler_func(
        total_min_delta, total_max_delta,
        batch_size, delta_vector.shape[1],
        device=total_min_bounds.device, dtype=total_min_bounds.dtype
    )

    prior_centers = uniform_sampler(
        total_min_bounds + prior_widths / 2, total_max_bounds - prior_widths / 2,
        *prior_widths.shape,
        device=total_min_bounds.device, dtype=total_min_bounds.dtype
    )

    min_bounds, max_bounds = prior_centers - prior_widths / 2, prior_centers + prior_widths / 2

    params = torch.rand(
        *min_bounds.shape,
        device=min_bounds.device,
        dtype=min_bounds.dtype
    ) * (max_bounds - min_bounds) + min_bounds

    return params, min_bounds, max_bounds


def constrained_roughness_sampler(
        batch_size: int,
        total_min_bounds: Tensor,
        total_max_bounds: Tensor,
        total_min_delta: Tensor,
        total_max_delta: Tensor,
        thickness_mask: Tensor,
        roughness_mask: Tensor,
        widths_sampler_func,
        coef_roughness: float = 0.5,
        max_total_thickness: float = None,
):
    params, min_bounds, max_bounds = basic_sampler(
        batch_size, total_min_bounds, total_max_bounds, total_min_delta, total_max_delta,
        widths_sampler_func=widths_sampler_func,
    )

    if max_total_thickness is not None:
        total_thickness = max_bounds[:, thickness_mask].sum(-1)
        indices = total_thickness > max_total_thickness

        if indices.any():
            eps = 0.01
            rand_scale = torch.rand_like(total_thickness) * eps + 1 - eps
            scale_coef = max_total_thickness / total_thickness * rand_scale
            scale_coef[~indices] = 1.0
            min_bounds[:, thickness_mask] *= scale_coef[:, None]
            max_bounds[:, thickness_mask] *= scale_coef[:, None]
            params[:, thickness_mask] *= scale_coef[:, None]

            min_bounds[:, thickness_mask] = torch.clamp_min(
                min_bounds[:, thickness_mask],
                total_min_bounds[:, thickness_mask],
            )

            max_bounds[:, thickness_mask] = torch.clamp_min(
                max_bounds[:, thickness_mask],
                total_min_bounds[:, thickness_mask],
            )

            params[:, thickness_mask] = torch.clamp_min(
                params[:, thickness_mask],
                total_min_bounds[:, thickness_mask],
            )

    max_roughness = torch.minimum(
        get_max_allowed_roughness(thicknesses=params[..., thickness_mask], coef=coef_roughness),
        total_max_bounds[..., roughness_mask]
    )
    min_roughness = total_min_bounds[..., roughness_mask]

    assert torch.all(min_roughness <= max_roughness)

    min_roughness_delta = total_min_delta[..., roughness_mask]
    max_roughness_delta = torch.minimum(total_max_delta[..., roughness_mask], max_roughness - min_roughness)

    roughnesses, min_r_bounds, max_r_bounds = basic_sampler(
        batch_size, min_roughness, max_roughness,
        min_roughness_delta, max_roughness_delta,
        widths_sampler_func=widths_sampler_func
    )

    min_bounds[..., roughness_mask], max_bounds[..., roughness_mask] = min_r_bounds, max_r_bounds
    params[..., roughness_mask] = roughnesses

    return params, min_bounds, max_bounds

def constrained_roughness_and_isld_sampler(
        batch_size: int,
        total_min_bounds: Tensor,
        total_max_bounds: Tensor,
        total_min_delta: Tensor,
        total_max_delta: Tensor,
        thickness_mask: Tensor,
        roughness_mask: Tensor,
        sld_mask: Tensor,
        isld_mask: Tensor,
        widths_sampler_func,
        coef_roughness: float = 0.5,
        coef_isld: float = 0.2,
        max_total_thickness: float = None,
):
    params, min_bounds, max_bounds = basic_sampler(
        batch_size, total_min_bounds, total_max_bounds, total_min_delta, total_max_delta,
        widths_sampler_func=widths_sampler_func,
    )

    if max_total_thickness is not None:
        total_thickness = max_bounds[:, thickness_mask].sum(-1)
        indices = total_thickness > max_total_thickness

        if indices.any():
            eps = 0.01
            rand_scale = torch.rand_like(total_thickness) * eps + 1 - eps
            scale_coef = max_total_thickness / total_thickness * rand_scale
            scale_coef[~indices] = 1.0
            min_bounds[:, thickness_mask] *= scale_coef[:, None]
            max_bounds[:, thickness_mask] *= scale_coef[:, None]
            params[:, thickness_mask] *= scale_coef[:, None]

            min_bounds[:, thickness_mask] = torch.clamp_min(
                min_bounds[:, thickness_mask],
                total_min_bounds[:, thickness_mask],
            )

            max_bounds[:, thickness_mask] = torch.clamp_min(
                max_bounds[:, thickness_mask],
                total_min_bounds[:, thickness_mask],
            )

            params[:, thickness_mask] = torch.clamp_min(
                params[:, thickness_mask],
                total_min_bounds[:, thickness_mask],
            )

    max_roughness = torch.minimum(
        get_max_allowed_roughness(thicknesses=params[..., thickness_mask], coef=coef_roughness),
        total_max_bounds[..., roughness_mask]
    )
    min_roughness = total_min_bounds[..., roughness_mask]

    assert torch.all(min_roughness <= max_roughness)

    min_roughness_delta = total_min_delta[..., roughness_mask]
    max_roughness_delta = torch.minimum(total_max_delta[..., roughness_mask], max_roughness - min_roughness)

    roughnesses, min_r_bounds, max_r_bounds = basic_sampler(
        batch_size, min_roughness, max_roughness,
        min_roughness_delta, max_roughness_delta,
        widths_sampler_func=widths_sampler_func
    )

    min_bounds[..., roughness_mask], max_bounds[..., roughness_mask] = min_r_bounds, max_r_bounds
    params[..., roughness_mask] = roughnesses

    max_isld = torch.minimum(
        torch.abs(params[..., sld_mask]) * coef_isld,
        total_max_bounds[..., isld_mask]
    )
    min_isld = total_min_bounds[..., isld_mask]

    assert torch.all(min_isld <= max_isld)

    min_isld_delta = total_min_delta[..., isld_mask]
    max_isld_delta = torch.minimum(total_max_delta[..., isld_mask], max_isld - min_isld)

    islds, min_isld_bounds, max_isld_bounds = basic_sampler(
        batch_size, min_isld, max_isld,
        min_isld_delta, max_isld_delta,
        widths_sampler_func=widths_sampler_func
    )

    min_bounds[..., isld_mask], max_bounds[..., isld_mask] = min_isld_bounds, max_isld_bounds
    params[..., isld_mask] = islds


    return params, min_bounds, max_bounds


from typing import Tuple

import torch
from torch import Tensor

from reflectorch.data_generation.utils import (
    get_d_rhos,
    uniform_sampler,
)


def get_max_allowed_roughness(thicknesses: Tensor, mask: Tensor = None, coef: float = 0.5):
    """gets the maximum allowed interlayer roughnesses such that they do not exceed a fraction of the thickness of either layers meeting at that interface"""
    batch_size, layers_num = thicknesses.shape
    max_roughness = torch.ones(
        batch_size, layers_num + 1, device=thicknesses.device, dtype=thicknesses.dtype
    ) * float('inf')

    boundary = thicknesses * coef
    if mask is not None:
        boundary[get_thickness_mask_from_sld_mask(mask)] = float('inf')

    max_roughness[:, :-1] = boundary
    max_roughness[:, 1:] = torch.minimum(max_roughness[:, 1:], boundary)
    return max_roughness


def get_allowed_contrast_indices(slds: Tensor, min_contrast: float, mask: Tensor = None) -> Tensor:
    d_rhos = get_d_rhos(slds)
    indices = d_rhos.abs() >= min_contrast
    if mask is not None:
        indices = indices | mask
    indices = torch.all(indices, -1)
    return indices


def params_within_bounds(params_t: Tensor, min_t: Tensor, max_t: Tensor, mask: Tensor = None) -> Tensor:
    indices = (params_t >= min_t[None]) & (params_t <= max_t[None])
    if mask is not None:
        indices = indices | mask
    indices = torch.all(indices, -1)
    return indices


def get_allowed_roughness_indices(thicknesses: Tensor, roughnesses: Tensor, mask: Tensor = None) -> Tensor:
    max_roughness = get_max_allowed_roughness(thicknesses, mask)
    indices = roughnesses <= max_roughness
    if mask is not None:
        indices = indices | mask
    indices = torch.all(indices, -1)
    return indices


def get_thickness_mask_from_sld_mask(mask: Tensor):
    return mask[:, :-1]


def generate_roughnesses(thicknesses: Tensor, roughness_range: Tuple[float, float], mask: Tensor = None):
    batch_size, layers_num = thicknesses.shape
    max_roughness = get_max_allowed_roughness(thicknesses, mask)
    max_roughness = torch.clamp_(max_roughness, max=roughness_range[1])

    roughnesses = uniform_sampler(
        roughness_range[0], max_roughness, batch_size, layers_num + 1,
        device=thicknesses.device, dtype=thicknesses.dtype
    )

    if mask is not None:
        roughnesses[mask] = 0.

    return roughnesses


def generate_thicknesses(
        thickness_range: Tuple[float, float],
        batch_size: int,
        layers_num: int,
        device: torch.device,
        dtype: torch.dtype,
        mask: Tensor = None
):
    thicknesses = uniform_sampler(
        *thickness_range, batch_size, layers_num, device=device, dtype=dtype
    )
    if mask is not None:
        thicknesses[get_thickness_mask_from_sld_mask(mask)] = 0.
    return thicknesses


def generate_slds_with_min_contrast(
        sld_range: Tuple[float, float],
        batch_size: int,
        layers_num: int,
        min_contrast: float,
        device: torch.device,
        dtype: torch.dtype,
        mask: Tensor = None,
        *,
        _depth: int = 0
):
    # rejection sampling
    slds = uniform_sampler(
        *sld_range, batch_size, layers_num + 1, device=device, dtype=dtype
    )

    if mask is not None:
        slds[mask] = 0.

    rejected_indices = ~get_allowed_contrast_indices(slds, min_contrast, mask)
    rejected_num = rejected_indices.sum(0).item()

    if rejected_num:
        if mask is not None:
            mask = mask[rejected_indices]
        slds[rejected_indices] = generate_slds_with_min_contrast(
            sld_range, rejected_num, layers_num, min_contrast, device, dtype, mask, _depth=_depth+1
        )
    return slds



from typing import Tuple

import torch
from torch import Tensor


class ScalerMixin:
    """Provides functionality to multiple inheritance classes for scaling the parameters to a specified range and restoring them to the original range."""
    @staticmethod
    def _get_delta_vector(min_vector: Tensor, max_vector: Tensor):
        delta_vector = max_vector - min_vector
        delta_vector[delta_vector == 0.] = 1.
        return delta_vector

    def _scale(self, params_t: Tensor, min_vector: Tensor, max_vector: Tensor):
        """scale the parameters to a specific range
        Args:
            params_t (Tensor): the values of the parameters
            min_vector (Tensor): minimum possible values of each parameter
            max_vector (Tensor): maximum possible values of each parameter

        Returns:
            Tensor: the scaled parameters
        """
        if params_t.dim() == 2:
            min_vector = torch.atleast_2d(min_vector)
            max_vector = torch.atleast_2d(max_vector)

        delta_vector = max_vector - min_vector
        delta_vector[delta_vector == 0.] = 1.
        scaled_params = (
                                params_t - min_vector
                        ) / self._get_delta_vector(min_vector, max_vector) * self._length + self._bias
        return scaled_params

    def _restore(self, scaled_params: Tensor, min_vector: Tensor, max_vector: Tensor):
        """restores the parameters to their original range
        Args:
            scaled_params: (Tensor): the scaled parameters
            min_vector (Tensor): minimum possible values of each parameter
            max_vector (Tensor): maximum possible values of each parameter

        Returns:
            Tensor: the restored parameters
        """
        if scaled_params.dim() == 2:
            min_vector = torch.atleast_2d(min_vector)
            max_vector = torch.atleast_2d(max_vector)

        params_t = (
                           scaled_params - self._bias
                   ) / self._length * self._get_delta_vector(min_vector, max_vector) + min_vector
        return params_t

    @property
    def scaled_range(self) -> Tuple[float, float]:
        return self._scaled_range

    @scaled_range.setter
    def scaled_range(self, scaled_range: Tuple[float, float]):
        """sets the range used for scaling the parameters"""
        self._scaled_range = scaled_range
        self._length = scaled_range[1] - scaled_range[0]
        self._init_bias = (scaled_range[0] + scaled_range[1]) / 2
        self._bias = (self._init_bias - 0.5 * self._length)



import torch
from torch import Tensor

from reflectorch.data_generation.priors.utils import uniform_sampler
from reflectorch.data_generation.priors.subprior_sampler import UniformSubPriorParams
from reflectorch.data_generation.priors.params import Params
from reflectorch.data_generation.likelihoods import LogLikelihood


def simple_sampler_solution(
        likelihood: LogLikelihood,
        predicted_params: UniformSubPriorParams,
        total_min_bounds: Tensor,
        total_max_bounds: Tensor,
        num: int = 2 ** 15,
        coef: float = 0.1,
) -> UniformSubPriorParams:
    sampled_params_t = sample_around_params(predicted_params, total_min_bounds, total_max_bounds, num=num, coef=coef)
    sampled_params = Params.from_tensor(sampled_params_t)
    return get_best_mse_param(sampled_params, likelihood, predicted_params.min_bounds, predicted_params.max_bounds)


def sample_around_params(predicted_params: UniformSubPriorParams,
                         total_min_bounds: Tensor,
                         total_max_bounds: Tensor,
                         num: int = 2 ** 15,
                         coef: float = 0.1,
                         ) -> Tensor:
    params_t = predicted_params.as_tensor(add_bounds=False)

    delta = (predicted_params.max_bounds - predicted_params.min_bounds) * coef
    min_bounds = torch.clamp(params_t - delta, total_min_bounds, total_max_bounds)
    max_bounds = torch.clamp(params_t + delta, total_min_bounds, total_max_bounds)

    sampled_params_t = uniform_sampler(min_bounds, max_bounds, num, params_t.shape[-1])
    sampled_params_t[0] = params_t[0]

    return sampled_params_t


def get_best_mse_param(
        params: Params,
        likelihood: LogLikelihood,
        min_bounds: Tensor = None,
        max_bounds: Tensor = None,
):
    sampled_curves = params.reflectivity(likelihood.q)
    log_probs = likelihood.calc_log_likelihood(sampled_curves)
    best_idx = torch.argmax(log_probs)
    best_param = params[best_idx:best_idx + 1]

    if min_bounds is not None:
        best_param = UniformSubPriorParams.from_tensor(
            torch.cat([best_param.as_tensor(), torch.atleast_2d(min_bounds), torch.atleast_2d(max_bounds)], -1)
        )
    return best_param
